from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from PySide6.QtCore import QItemSelectionModel, QModelIndex, QObject, Qt, SignalInstance
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QLineEdit,
    QTextBrowser,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

import conan_explorer.app as app
from conan_explorer import asset_path
from conan_explorer.app import LoaderGui  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanPkg, ConanPkgRef, ConanRef
from conan_explorer.ui.common import ConfigHighlighter, show_conanfile
from conan_explorer.ui.dialogs import ConanInstallDialog, ConanRemoveDialog
from conan_explorer.ui.dialogs.pkg_diff.diff import PkgDiffDialog

from .sel_model import PackageFilter, PackageTreeItem, PkgSelectionType, PkgSelectModel

if TYPE_CHECKING:
    from conan_explorer.ui.fluent_window import FluentWindow
    from conan_explorer.ui.main_window import BaseSignals


class MultiPkgSelectionMode(Enum):
    single_ref = 0
    single_pkg_or_export = 1
    multi_ref = 2
    multi_pkg = 3
    single_editable = 4
    invalid = -1


class PackageSelectionController(QObject):
    def __init__(
        self,
        parent: QWidget,
        view: QTreeView,
        package_filter_edit: QLineEdit,
        conan_pkg_selected: SignalInstance,
        base_signals: "BaseSignals",
        page_widgets: "FluentWindow.PageStore",
    ):
        super().__init__(parent)
        self._base_signals = base_signals
        self._conan_pkg_selected = conan_pkg_selected
        self._model = None
        self._loader = LoaderGui(self)
        self._page_widgets = page_widgets
        self._view = view
        self._package_filter_edit = package_filter_edit

        base_signals.conan_pkg_removed.connect(self.on_conan_pkg_removed)

    # Own slots

    def on_pkg_refresh_clicked(self):
        self.refresh_pkg_selection_view()

    def on_show_conanfile_requested(self):
        conan_refs = self.get_selected_conan_refs()
        if len(conan_refs) != 1:
            return
        loader = LoaderGui(self)
        loader.load(
            self._view, show_conanfile, (conan_refs[0],), loading_text="Opening Conanfile..."
        )

    def on_copy_ref_requested(self):
        conan_refs = self.get_selected_conan_refs()
        QApplication.clipboard().setText("\n".join(conan_refs))

    def on_install_ref_requested(self):
        conan_ref, pkg_id = self.get_selected_ref_with_pkg_id()
        if not conan_ref:
            return
        if pkg_id:
            conan_ref += ":" + pkg_id
        dialog = ConanInstallDialog(
            self._view, conan_ref, self._base_signals.conan_pkg_installed, lock_reference=True
        )
        dialog.show()

    def on_remove_ref_requested(self):
        conan_refs_with_pkg_ids = self.get_selected_refs_with_pkg_ids()
        if not conan_refs_with_pkg_ids:
            return
        dialog = ConanRemoveDialog(
            self._view, conan_refs_with_pkg_ids, self._base_signals.conan_pkg_removed
        )
        dialog.show()

    def on_conan_pkg_removed(self, conan_ref: str, pkg_id: str):
        """Remove conan reference if in model"""
        if not self._model:  # for None usage
            return
        ref_row = self.find_item_in_pkg_sel_model(conan_ref, pkg_id)
        if not ref_row:
            return
        proxy_index = self._model.index(ref_row, 0, QModelIndex())
        item: Optional[PackageTreeItem] = proxy_index.internalPointer()  # type: ignore
        if not item:
            return
        self._model.remove_item(item)

    def on_show_build_info(self):
        conan_ref, pkg_id = self.get_selected_ref_with_pkg_id()
        if not conan_ref or not pkg_id:
            return
        pkg_info = app.conan_api.get_local_pkg_from_id(
            ConanPkgRef.loads(conan_ref + ":" + pkg_id)
        )
        loader = LoaderGui(self)
        loader.load(
            self._view,
            app.conan_api.get_conan_buildinfo,
            (
                ConanRef.loads(conan_ref),
                pkg_info.get("settings", ""),
                pkg_info.get("options", {}),
            ),
            self.show_buildinfo_dialog,
            loading_text="Loading build info...",
        )

    def show_buildinfo_dialog(self, buildinfos: str):
        if not buildinfos:
            return
        dialog = QDialog(self._view)
        dialog.setWindowIcon(QIcon(str(asset_path / "icons" / "icon.ico")))
        dialog.setWindowTitle("Build Info")
        dialog.resize(800, 500)
        verticalLayout = QVBoxLayout(dialog)
        text_browser = QTextBrowser(dialog)
        text_browser.setReadOnly(True)
        text_browser.setOpenExternalLinks(True)
        text_browser.setText(buildinfos)
        verticalLayout.addWidget(text_browser)
        self.conan_config_highlighter = ConfigHighlighter(text_browser.document(), "ini")
        dialog.exec()

    def on_diff_requested(self):
        dialog = PkgDiffDialog(self._view)
        items = self.get_selected_pkg_source_items()
        if len(items) < 2:
            return
        for item in items:
            dialog.add_diff_item(item.pkg_info)
        dialog.show()

    # Model selection helpers

    def get_selected_pkg_source_items(self) -> List[PackageTreeItem]:
        indexes = self._view.selectedIndexes()
        if len(indexes) == 0:
            return []
        model: PackageFilter = indexes[0].model()  # type: ignore
        source_items: List[PackageTreeItem] = []
        for index in indexes:
            source_items.append(model.mapToSource(index).internalPointer())  # type: ignore
        return list(set(source_items))

    def get_selected_ref_with_pkg_id(self) -> Tuple[str, str]:
        conan_refs = self.get_selected_conan_refs()
        if len(conan_refs) != 1:
            return "", ""
        pkg_info = self.get_selected_conan_pkg_info()
        pkg_id = ""
        if pkg_info:
            pkg_id = pkg_info.get("id", "")
        return conan_refs[0], pkg_id

    def get_selected_refs_with_pkg_ids(self) -> Dict[str, List[str]]:
        refs_with_pkg_ids = {}
        source_items = self.get_selected_pkg_source_items()
        for source_item in source_items:
            conan_ref_item = source_item
            if source_item.type in [PkgSelectionType.export, PkgSelectionType.editable]:
                continue
            if source_item.type == PkgSelectionType.pkg:
                conan_ref_item: PackageTreeItem = source_item.parent()  # type: ignore
            conan_ref: str = conan_ref_item.item_data[0]
            pkg_id = ""
            if pkg_info := source_item.pkg_info:
                pkg_id = pkg_info.get("id", "")
            if conan_ref not in refs_with_pkg_ids:
                refs_with_pkg_ids[conan_ref] = [pkg_id]
            else:
                refs_with_pkg_ids[conan_ref].append(pkg_id)

        # clean up dict -> empty str means the ref was selected -> delete all other ids
        for ref, pkg_ids in refs_with_pkg_ids.items():
            if "" in pkg_ids:
                refs_with_pkg_ids[ref] = [""]
        return refs_with_pkg_ids

    def get_selection_mode(self, source_items: List[PackageTreeItem]) -> MultiPkgSelectionMode:
        """ " Determine which combination of multiselected items is active"""
        pkg_or_export = [PkgSelectionType.pkg, PkgSelectionType.export]
        if len(source_items) == 1:
            if source_items[0].type == PkgSelectionType.ref:
                return MultiPkgSelectionMode.single_ref
            elif source_items[0].type in pkg_or_export:
                return MultiPkgSelectionMode.single_pkg_or_export
            elif source_items[0].type == PkgSelectionType.editable:
                return MultiPkgSelectionMode.single_editable
        if all([item.type == PkgSelectionType.ref for item in source_items]):
            return MultiPkgSelectionMode.multi_ref
        elif all([item.type == PkgSelectionType.pkg for item in source_items]):
            return MultiPkgSelectionMode.multi_pkg
        return MultiPkgSelectionMode.invalid

    def get_selected_conan_refs(self) -> List[str]:
        # no need to map from postition, since rightclick selects a single item
        source_items = self.get_selected_pkg_source_items()
        conan_refs = []
        for conan_ref_item in source_items:
            if conan_ref_item.type in [PkgSelectionType.pkg, PkgSelectionType.export]:
                conan_ref_item: PackageTreeItem = conan_ref_item.parent()  # type: ignore
            if not conan_ref_item:
                conan_refs.append("")
            conan_refs.append(conan_ref_item.item_data[0])
        return list(set(conan_refs))

    def get_selected_conan_pkg_info(self) -> ConanPkg:
        source_items = self.get_selected_pkg_source_items()
        if not len(source_items) == 1:
            return ConanPkg()
        source_item = source_items[0]
        if source_item.type == PkgSelectionType.ref:
            return ConanPkg()
        return source_item.pkg_info

    # Global pane and cross connection slots

    def on_pkg_selection_change(self):
        # get conan ref and pkg_id and emit signal
        source_items = self.get_selected_pkg_source_items()
        if len(source_items) != 1:
            return
        if source_items[0].type in [PkgSelectionType.ref]:
            return
        conan_refs = self.get_selected_conan_refs()
        self._conan_pkg_selected.emit(
            conan_refs[0], self.get_selected_conan_pkg_info(), source_items[0].type
        )

    def refresh_pkg_selection_view(self, force_update=True):
        """
        Refresh all packages by reading it from local drive. Can take a while.
        """
        if self._model and not force_update:  # loads only at first init
            return
        if not self._model:
            self._model = PkgSelectModel(self._loader.loading_string_signal)
        self._loader.load(
            self._view,
            self._model.setup_model_data,
            (),
            self.finish_select_model_init,
            "Reading Packages",
        )

    def finish_select_model_init(self):
        if self._model:
            self._view.setModel(self._model.proxy_model)
            self._view.selectionModel().selectionChanged.connect(self.on_pkg_selection_change)
            self.set_filter_wildcard()  # re-apply package filter query
            self._view.hideColumn(1)  # don't show size on opening view
        else:
            Logger().error("Can't load local packages!")

    def on_show_sizes(self):
        if not self._model:
            return
        if not self._model.show_sizes:
            self._model.show_sizes = True
            self._view.showColumn(1)
            self._loader.load(
                self._view,
                self._model.get_all_sizes,
                (),
                self.expand_and_sort_for_sizes,
                "Calculating Sizes. This can take a while...",
            )
        else:
            self._model.show_sizes = False
            self._view.collapseAll()
            self._view.hideColumn(1)

    def expand_and_sort_for_sizes(self):
        self._view.expandAll()
        self._view.sortByColumn(1, Qt.SortOrder.DescendingOrder)
        self._view.header().resizeSections(self._view.header().ResizeMode.Stretch)

    def set_filter_wildcard(self):
        # use strip to remove unnecessary whitespace
        text = self._package_filter_edit.text().strip()  # toPlainText
        if self._model:
            self._model.proxy_model.setFilterWildcard(text)

    def find_item_in_pkg_sel_model(self, conan_ref: str, pkg_id="") -> int:
        # find the row with the matching reference
        if not self._model:
            return False
        for ref_row in range(self._model.root_item.child_count()):
            try:  # try for empty model
                item = self._model.root_item.child_items[ref_row]
            except Exception:
                return -1
            if item.item_data[0] == conan_ref:
                if pkg_id:
                    if item.child_count() < 2:  # try to fetch, pkd_id probably not loaded yet
                        item.load_children()
                    for pkg_row in range(item.child_count()):
                        pkg_item: PackageTreeItem = item.child_items[pkg_row]  # type: ignore
                        if pkg_item.pkg_info.get("id") == pkg_id:
                            return ref_row
                    return -1
                else:
                    return ref_row
        return -1

    def select_local_package_from_ref(
        self,
        conan_ref: str,
        export=False,
        select_mode=QItemSelectionModel.SelectionFlag.ClearAndSelect,
    ) -> Optional[QModelIndex]:
        """Selects a reference:id pkg in the left pane and opens the file view
        param export: tells to select the export folder
        """
        # change to this page and loads
        self._page_widgets.get_button_by_type(type(self.parent())).click()
        self._loader.wait_for_finished()
        if not self._model:
            return None

        # Reset filter, otherwise the element to be shown could be hidden
        self._package_filter_edit.setText("*")

        # find out if we need to find a ref or or a package
        error_message_suffix = " in Local Package Explorer for selection."
        split_ref = conan_ref.split(":")
        pkg_id = ""
        if len(split_ref) > 1:  # has id
            if export:
                Logger().debug(
                    "Cannot use pkg id and export arg at the same time" + error_message_suffix
                )
                return None
            conan_ref = split_ref[0]
            pkg_id = split_ref[1]

        # not found ref or id - start refresh package list
        if self.find_item_in_pkg_sel_model(conan_ref, pkg_id) == -1:
            self.refresh_pkg_selection_view()

        # wait for model to be loaded
        self._loader.wait_for_finished()
        ref_row = self.find_item_in_pkg_sel_model(conan_ref, pkg_id)
        if ref_row == -1:
            Logger().debug("Cannot find %s%s", conan_ref, error_message_suffix)
            return None
        Logger().debug("Found %s@%d%s", conan_ref, ref_row, error_message_suffix)

        # map to package view model
        proxy_index = self._model.index(ref_row, 0, QModelIndex())
        sel_model = self._view.selectionModel()
        view_model: PackageFilter = self._view.model()  # type: ignore
        self._view.expand(view_model.mapFromSource(proxy_index))

        # retrieve item from id, export or ref
        if pkg_id:
            item: PackageTreeItem = proxy_index.internalPointer()  # type: ignore
            i = 0
            for i, child_item in enumerate(item.child_items):
                child_item: PackageTreeItem
                if child_item.type == PkgSelectionType.pkg:
                    if child_item.pkg_info.get("id", "") == pkg_id:
                        break
            internal_sel_index = self._model.index(i, 0, proxy_index)
        else:
            if export:
                item: PackageTreeItem = proxy_index.internalPointer()  # type: ignore
                i = 0
                for i, child_item in enumerate(item.child_items):
                    if child_item.type == PkgSelectionType.export:
                        break
                internal_sel_index = self._model.index(i, 0, proxy_index)
            else:
                internal_sel_index = proxy_index

        # Scroll to item and select
        view_index = view_model.mapFromSource(internal_sel_index)
        self._view.scrollTo(view_index)
        sel_model.select(view_index, select_mode)
        sel_model.currentRowChanged.emit(proxy_index, internal_sel_index)
        Logger().debug("Selecting %s in Local Package Explorer", str(view_index.data()))

        return view_index



from typing import List, Optional

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QFrame,
    QLayout,
    QScrollArea,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from .app_link import ListAppLink  # ,GridAppLink
from .dialogs import AppEditDialog
from .model import UiAppLinkModel, UiTabModel


class TabScrollAreaWidgets(QWidget):
    def __init__(self, parent: Optional["QWidget"] = None):
        super().__init__(parent)


class TabList(QWidget):
    SPACING = 10

    def __init__(self, parent: QTabWidget, model: UiTabModel):
        super().__init__(parent)
        self.model = model
        self._parent_tab = parent
        self.app_links: List[ListAppLink] = []  # list of refs to app links
        self._initialized = False
        self._columns_count = 0
        self._v_spacer = QSpacerItem(
            20, 2000, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )

    def load(self, offset=0):
        self.init_app_grid()
        self.load_apps_from_model(offset=offset)

    def load_apps_from_model(self, force_reload=False, offset=0):
        if not self.app_links or force_reload:
            for app_model in self.model.apps:
                app_link = ListAppLink(self, self, app_model)
                app_link.load()
                self.app_links.append(app_link)
        for app_link in self.app_links:
            # add in order of occurence
            self.tab_layout.addWidget(app_link)
            app_link.show()
        # spacer for compressing app links, when hiding cboxes
        self.tab_layout.addItem(self._v_spacer)

    def open_app_link_add_dialog(self, new_model: Optional[UiAppLinkModel] = None):
        if not new_model:
            new_model = UiAppLinkModel()
            new_model.parent = self.model
        # save for testing
        self._edit_app_dialog = AppEditDialog(new_model, parent=self)
        reply = self._edit_app_dialog.exec()
        if reply == AppEditDialog.DialogCode.Accepted:
            self.model.apps.append(new_model)
            self.model.save()
            self.redraw(force=True)
            return new_model  # for testing
        return None

    def remove_all_app_links(self, force=False):
        """
        Clears all AppLinks.
        Can then be reloaded with load_apps_from_model.
        """
        # remove spacer - needed so, the layout can be resized correctly, if layout shifts
        self.tab_layout.removeItem(self._v_spacer)
        # self.tab_layout.removeItem(self._h_spacer)
        for app_link in self.app_links:
            self.tab_layout.removeWidget(app_link)
            # TODO this a leak, currently it does not delete everything!
            app_link.hide()
        if force:
            self.app_links = []

    def redraw(self, force=False):
        if self._initialized:  # don't call on init
            self.remove_all_app_links(force)
            self.load_apps_from_model(force)

    def init_app_grid(self):
        self.setObjectName("tab_" + self.model.name)
        self.setContentsMargins(0, 0, 0, 0)
        # this is a dummy, because tab_scroll_area needs a layout
        self.tab_layout = QVBoxLayout(self)
        self.tab_layout.setContentsMargins(0, 0, 0, 0)
        self.tab_layout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.setLayout(self.tab_layout)

        # makes it possible to have a scroll bar
        self.tab_scroll_area = QScrollArea(self)
        self.tab_scroll_area.setContentsMargins(0, 0, 0, 0)
        self.tab_scroll_area.setFrameStyle(QFrame.Shape.NoFrame)

        self.tab_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.tab_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.tab_scroll_area.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop
        )
        self.tab_scroll_area.setWidgetResizable(True)
        # this holds all the app links, which are layouts
        self.tab_scroll_area_widgets = TabScrollAreaWidgets(self.tab_scroll_area)
        self.tab_scroll_area_widgets.setContentsMargins(0, 0, 0, 0)
        self.tab_scroll_area_widgets.setObjectName("tab_widgets_" + self.model.name)
        self._initialized = True

        self.tab_layout = QVBoxLayout(self.tab_scroll_area_widgets)
        size_policy = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding
        )
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(0)

        self.tab_scroll_area_widgets.setSizePolicy(size_policy)
        self.tab_scroll_area_widgets.setLayoutDirection(Qt.LayoutDirection.LeftToRight)
        self.tab_layout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.tab_layout.setSpacing(self.SPACING)
        self.tab_layout.setContentsMargins(0, 0, 5, 0)

        self.tab_scroll_area.setWidget(self.tab_scroll_area_widgets)
        self.layout().addWidget(self.tab_scroll_area)

    def add_app_link_to_tab(self, app_link: ListAppLink):
        """To be called from a child AppLink"""
        self.app_links.append(app_link)
        self.model.apps.append(app_link.model)
        self.tab_layout.insertWidget(-1, app_link)  # -1 is end
        self.tab_layout.update()



from typing import List, Optional

from PySide6.QtCore import QAbstractListModel, QModelIndex, Qt
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer.ui.common import get_platform_icon


class ProfilesModel(QAbstractListModel):
    def __init__(self):
        super().__init__()
        self._profiles: List[str] = []
        self.setup_model_data()

    def setup_model_data(self):
        self._profiles = app.conan_api.get_profiles()

    @override
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            text = self._profiles[index.row()]
            return text
        # platform logo
        if role == Qt.ItemDataRole.DecorationRole:
            text = self._profiles[index.row()]
            return get_platform_icon(text)

    @override
    def rowCount(self, parent):
        return len(self._profiles)

    def get_index_from_profile(self, profile_name: str) -> Optional[QModelIndex]:
        index = None
        for i, profile in enumerate(self._profiles):
            if profile == profile_name:
                index = self.createIndex(i, 0)
                break
        return index



from contextlib import redirect_stderr, redirect_stdout
import os
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union
from typing_extensions import Self
from unittest.mock import patch

from conan_explorer import INVALID_PATH, user_save_path, conan_version, Version
from conan_explorer.app.logger import Logger
from conan_explorer.app.typing import SignatureCheckMeta

from .types import (ConanAvailableOptions, ConanException, ConanOptions, ConanPackageId,
    ConanPackagePath, ConanPkg, ConanPkgRef, ConanRef, ConanSettings, EditablePkg, Remote,
    create_key_value_pair_list)
from .unified_api import ConanCommonUnifiedApi

if TYPE_CHECKING:
    from conan.api.conan_api import ConanAPI # type: ignore
    from conan.internal.cache.cache import PkgCache as ClientCache  # type: ignore
    from .conan_cache import ConanInfoCache
    from conan.internal.cache.home_paths import HomePaths  # type: ignore

class ConanApi(ConanCommonUnifiedApi, metaclass=SignatureCheckMeta):
    """ Wrapper around ConanAPIV2 """

    def __init__(self):
        self.info_cache: "ConanInfoCache"
        self._conan: "ConanAPI"
        self._client_cache: "ClientCache"
        self._short_path_root = Path("Unknown")
        self._home_paths: "HomePaths"

    def init_api(self) -> Self:
        from conan.api.conan_api import ConanAPI
        devnull = open(os.devnull, 'w')
        with redirect_stdout(devnull), redirect_stderr(devnull):
            self._conan = ConanAPI()
            self.init_client_cache()

        from .conan_cache import ConanInfoCache
        self.info_cache = ConanInfoCache(user_save_path, self.get_all_local_refs())
        Logger().debug("Initialized Conan V2 API wrapper")
        return self

    def init_client_cache(self):
        if conan_version < Version("2.4.0"):
            from conans.client.cache.cache import ClientCache
        else:
            from conan.internal.cache.cache import PkgCache as ClientCache
        self._client_cache = ClientCache(self._conan.cache_folder, self._conan.config.global_conf)
        from conan.internal.cache.home_paths import HomePaths
        self._home_paths = HomePaths(self._conan.cache_folder)


    ### General commands ###


    def remove_locks(self):
        pass  # command does not exist

    def get_profiles(self) -> List[str]:
        return self._conan.profiles.list()

    def get_profile_settings(self, profile_name: str) -> ConanSettings:
        try:
            from conans.client.profile_loader import ProfileLoader
        except:
            from conan.internal.api.profile.profile_loader import ProfileLoader

        try:
            profile = ProfileLoader(self._conan.cache_folder).load_profile(profile_name)
            return profile.settings
        except Exception as e:
            Logger().error(f"Can't get profile {profile_name} settings: {str(e)}")
        return {}

    def get_package_folder(self, conan_ref: ConanRef, package_id: str) -> Path:
        if not package_id:  # will give the base path ortherwise
            return Path(INVALID_PATH)
        try:
            latest_rev_ref = self._conan.list.latest_recipe_revision(conan_ref)
            latest_rev_pkg = self._conan.list.latest_package_revision(ConanPkgRef(latest_rev_ref, package_id))
            assert latest_rev_pkg
            layout = self._client_cache.pkg_layout(latest_rev_pkg)
            return Path(layout.package())
        except Exception:  # gotta catch 'em all!
            return Path(INVALID_PATH)

    def get_export_folder(self, conan_ref: ConanRef) -> Path:
        return Path(self._conan.cache.export_path(conan_ref))

    def get_conanfile_path(self, conan_ref: ConanRef) -> Path:
        try:
            if conan_ref not in self.get_all_local_refs():
                for remote in self.get_remotes():
                    result = self.search_recipes_in_remotes(str(conan_ref), remote_name=remote.name)
                    if result:
                        latest_rev: ConanRef = self._conan.list.latest_recipe_revision(conan_ref, remote)
                        self._conan.download.recipe(latest_rev, remote) # type: ignore
                        break
            path = self._conan.local.get_conanfile_path(
                self._conan.cache.export_path(conan_ref), os.getcwd(), py=True)
            if not path:
                return Path(INVALID_PATH)
            return Path(path)
        except Exception as e:
            Logger().error(f"Can't get conanfile: {str(e)}")
        return Path(INVALID_PATH)

    def get_default_settings(self) -> ConanSettings:
        try:
            from conans.client.profile_loader import ProfileLoader
        except:
            from conan.internal.api.profile.profile_loader import ProfileLoader
        profile = ProfileLoader(self._conan.cache_folder).load_profile(
                Path(self._conan.profiles.get_default_host()).name)
        
        return dict(profile.settings)

    # user_name, authenticated
    def get_remote_user_info(self, remote_name: str) -> Tuple[str, bool]:
        try:
            info = self._conan.remotes.user_info(
                self._conan.remotes.get(remote_name))
        except Exception as e:
            Logger().error(
                f"Can't get remote {remote_name} user info: {str(e)}")
            return "", False
        return str(info.get("user_name", "")), info.get("authenticated", False)

    def get_config_file_path(self) -> Path:
        if conan_version < Version("2.0.14"):
            cf_path = Path(self._client_cache.new_config_path)
        else:
            cf_path = Path(self._home_paths.global_conf_path)
        return cf_path

    def get_config_entry(self, config_name: str, default_value: Any) -> Any:
        return self._conan.config.get(config_name, default_value)

    def get_revisions_enabled(self) -> bool:
        return True # always on in 2

    def get_settings_file_path(self) -> Path:
        settings_path = Path(self._home_paths.settings_path)
        return settings_path

    def get_profiles_path(self) -> Path:
        profiles_path = Path(self._home_paths.profiles_path)
        return profiles_path

    def get_user_home_path(self) -> Path:
        return Path(self._conan.cache_folder)

    def get_storage_path(self) -> Path:
        return Path(str(self._client_cache.store))

    def get_short_path_root(self) -> Path:
        # there is no short_paths feature in conan 2
        return Path(INVALID_PATH)

    # Remotes

    def get_remotes(self, include_disabled=False) -> List[Remote]:
        remotes = []
        try:
            remotes = self._conan.remotes.list(None, only_enabled=not include_disabled)
        except Exception as e:
            Logger().error(f"Error while reading remotes: {str(e)}")
        return remotes # type: ignore

    def add_remote(self, remote_name: str, url: str, verify_ssl: bool):
        if conan_version < Version("2.1"):
            from conans.client.cache.remote_registry import Remote as ConanRemote
        else:
            from conan.api.model import Remote as ConanRemote
        remote = ConanRemote(remote_name, url, verify_ssl, False)
        self._conan.remotes.add(remote)

    def rename_remote(self, remote_name: str, new_name: str):
        self._conan.remotes.rename(remote_name, new_name)

    def remove_remote(self, remote_name: str):
        self._conan.remotes.remove(remote_name)

    def disable_remote(self, remote_name: str, disabled: bool):
        if disabled:
            self._conan.remotes.disable(remote_name)
        else:
            self._conan.remotes.enable(remote_name)

    def update_remote(self, remote_name: str, url: str, verify_ssl: bool, disabled: bool,
                      index: Optional[int]):
        self._conan.remotes.update(remote_name, url, verify_ssl,
                                   self._conan.remotes.get(remote_name).disabled, index)

    def login_remote(self, remote_name: str, user_name: str, password: str):
        if conan_version < Version("2.1"):
            self._conan.remotes.login(self._conan.remotes.get(
                remote_name), user_name, password)
        else:
            self._conan.remotes.user_login(self._conan.remotes.get(
                remote_name), user_name, password)
    ### Install related methods ###
    def install_reference(self, conan_ref: ConanRef, conan_settings: Optional[ConanSettings]=None,
            conan_options: Optional[ConanOptions]=None, profile="", update=True, quiet=False,
            generators: List[str] = []) -> Tuple[ConanPackageId, ConanPackagePath]:
        pkg_id = ""
        if conan_options is None:
            conan_options = {}
        if conan_settings is None:
            conan_settings = {}
        options_list = create_key_value_pair_list(conan_options)
        settings_list = create_key_value_pair_list(conan_settings)
        if not quiet:
            install_message = f"Installing '<b>{str(conan_ref)}</b>' with profile: {profile}, " \
                f"settings: {str(settings_list)}, " \
                f"options: {str(options_list)} and update={update}\n"
            Logger().info(install_message)
        from conan.cli.printers.graph import print_graph_basic, print_graph_packages

        try:
            # Basic collaborators, remotes, lockfile, profiles
            remotes = self._conan.remotes.list(None)
            profiles = [profile] if profile else []
            profile_host = self._conan.profiles.get_profile(profiles, 
                                        settings=settings_list, options=options_list)
            requires = [conan_ref]
            if conan_version > Version("2.0"):
                update = "*" if update else ""
            deps_graph = self._conan.graph.load_graph_requires(requires, None, 
                            profile_host, profile_host, None, remotes, update)
            print_graph_basic(deps_graph)
            deps_graph.report_graph_error()
            self._conan.graph.analyze_binaries(deps_graph, build_mode=None, remotes=remotes, 
                                               update=update, lockfile=None)
            print_graph_packages(deps_graph)

            # Try to redirect custom streams in conanfile, to avoid missing flush method
            devnull = open(os.devnull, 'w')
            # also spoof os.terminal_size(
            spoof_size = os.terminal_size([80,20])
            patched_tersize = patch("os.get_terminal_size")
            with redirect_stdout(devnull), redirect_stderr(devnull):
                mock = patched_tersize.start()
                mock.return_value = spoof_size

                self._conan.install.install_binaries(deps_graph=deps_graph, remotes=remotes)
                # Currently unused
                self._conan.install.install_consumer(deps_graph=deps_graph, 
                    generators=generators, output_folder=None, source_folder=gettempdir())

                patched_tersize.stop()

            info = None
            for node in deps_graph.nodes:
                if node.ref == conan_ref:
                    info = node
                    break
            if info is None:
                raise ConanException(
                    "Can't read information of installed recipe from graph.")
            pkg_id = info.package_id
            Logger().info(
                f"Installation of '<b>{str(conan_ref)}</b>' finished")
            # Update cache with this package
            self.info_cache.update_local_package_path(
                conan_ref, self.get_package_folder(conan_ref, pkg_id))
            return (pkg_id, self.get_package_folder(conan_ref, pkg_id))
        except ConanException as error:
            Logger().error(
                f"Can't install reference '<b>{str(conan_ref)}</b>': {str(error)}")
            return (pkg_id, Path(INVALID_PATH))

    def get_options_with_default_values(self, 
                    conan_ref: ConanRef) -> Tuple[ConanAvailableOptions, ConanOptions]:
        # this calls external code of the recipe
        default_options = {}
        available_options = {}
        try:
            path = self.get_conanfile_path(conan_ref)
            from conan.internal.conan_app import ConanApp
            if conan_version < Version("2.1"):
                app = ConanApp(self._conan.cache_folder, self._conan.config.global_conf)
            else:
                app = ConanApp(self._conan)

            conanfile = app.loader.load_conanfile(path, conan_ref)
            default_options = {}
            if conanfile.default_options is not None:
                default_options = conanfile.default_options
            available_options = conanfile.options
            default_options = self._resolve_default_options(default_options)
        except Exception as e:  # silent error - if we have no options don't spam the user
            Logger().debug(
                f"Error while getting default options for {str(conan_ref)}: {str(e)}", exc_info=True)
        return available_options, default_options

    ### Local References and Packages ###

    def get_conan_buildinfo(self, conan_ref: ConanRef, conan_settings: ConanSettings,
                            conan_options: Optional[ConanOptions]=None) -> str:
        """ TODO: Currently there is no equivalent to txt generator from ConanV1 """
        raise NotImplementedError
    
    def get_editable(self, conan_ref: Union[ConanRef, str]) -> EditablePkg:
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        if isinstance(conan_ref, str):
            conan_ref = ConanRef.loads(conan_ref)
        editable_dict = editables_dict.get(conan_ref, {})
        return EditablePkg(str(conan_ref), editable_dict.get("path", INVALID_PATH), 
                           editable_dict.get("output_folder"))
    
    def get_editables_package_path(self, conan_ref: ConanRef) -> Path:
        """ Get package path of an editable reference. Can be a folder or conanfile.py """
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        return Path(editables_dict.get(conan_ref, {}).get("path", INVALID_PATH))
    
    def get_editables_output_folder(self, conan_ref: ConanRef) ->  Optional[Path]:
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        output_folder = editables_dict.get(conan_ref, {}).get("output_folder")
        if not output_folder:
            return None
        return Path(str(output_folder))

    def get_editable_references(self) -> List[ConanRef]:
        """ Get all local editable references. """
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        return list(editables_dict.keys())
    
    def __reload_conan_api(self):
        from conan.api.conan_api import ConanAPI
        self._conan = ConanAPI()  # reload editables only possible like this


    def add_editable(self, conan_ref: Union[ConanRef, str], path: str, output_folder: str) -> bool:
        try:
            if isinstance(conan_ref, str):
                conan_ref = ConanRef.loads(conan_ref)
            self._conan.local.editable_add(path, conan_ref.name, conan_ref.version,
                conan_ref.user, conan_ref.channel, output_folder=output_folder)
        except Exception as e:
            Logger().error("Error adding editable: " + str(e))
            return False
        return True

    def remove_editable(self, conan_ref: Union[ConanRef, str]) -> bool:
        try:
            if isinstance(conan_ref, str):
                conan_ref = ConanRef.loads(conan_ref)
            self._conan.local.editable_remove(None, [str(conan_ref)])
        except Exception as e:
            Logger().error("Error removing editable: " + str(e))
            return False
        return True

    def remove_reference(self, conan_ref: ConanRef, pkg_id: str = ""):
        if pkg_id:
            conan_pkg_ref = ConanPkgRef.loads(str(conan_ref) + ":" + pkg_id)
            latest_rev = self._conan.list.latest_package_revision(conan_pkg_ref)
            self._conan.remove.package(latest_rev, remote=None)  # type: ignore
        else:
            latest_rev = self._conan.list.latest_recipe_revision(conan_ref)
            self._conan.remove.recipe(latest_rev, remote=None)  # type: ignore

    def get_all_local_refs(self) -> List[ConanRef]:
        # Sadly not exposed anymore?!
        if conan_version < Version("2.6"):
            return self._client_cache.all_refs()
        return self._client_cache._db.list_references()

    def get_local_pkg_from_path(self, conan_ref: ConanRef, path: Path):
        found_package = None
        for package in self.get_local_pkgs_from_ref(conan_ref):
            if self.get_package_folder(conan_ref, package.get("id", "")) == path:
                found_package = package
                break
        return found_package

    def get_local_pkgs_from_ref(self, conan_ref: ConanRef) -> List[ConanPkg]:
        result: List[ConanPkg] = []
        if conan_ref.user == "_":
            conan_ref.user = None
        if conan_ref.channel == "_":
            conan_ref.channel = None
        if not conan_ref.revision:
            try:
                conan_ref_latest: ConanRef = self._conan.list.latest_recipe_revision(conan_ref)
            except Exception as e:
                Logger().error(
                    f"Error while getting latest recipe for {str(conan_ref)}: {str(e)}")
                return result
            if not conan_ref_latest:
                return result
        else:
            conan_ref_latest = conan_ref
        try:  # errors with invalid pkg
            refs = self._conan.list.packages_configurations(conan_ref_latest)
        except Exception as e:
            Logger().error(
                f"Error while getting packages for recipe {str(conan_ref)}: {str(e)}")
            return result
        for ref, pkg_info in refs.items():
            pkg = ConanPkg()
            pkg["id"] = str(ref.package_id)
            pkg["options"] = pkg_info.get("options", {})
            pkg["settings"] = pkg_info.get("settings", {})
            pkg["requires"] = []
            pkg["outdated"] = False
            result.append(pkg)
        return result

    ### Remote References and Packages ###

    def search_recipes_in_remotes(self, query: str, remote_name="all") -> List[ConanRef]:
        search_results = []
        if remote_name == "all":
            remote_name = None
        try:
            # no query possible with pattern
            for remote in self.get_remotes():
                search_results: List["ConanRef"] = self._conan.search.recipes(
                    query, remote=remote)
        except Exception as e:
            Logger().error(f"Error while searching for recipe: {str(e)}")
            return []

        search_results = list(set(search_results))  # make unique
        search_results.sort()
        return search_results

    def search_recipe_all_versions_in_remotes(self, conan_ref: ConanRef) -> List[ConanRef]:
        search_results = []
        try:
            # no query possible with pattern
            search_results: List = self.search_recipes_in_remotes(f"{conan_ref.name}/*@*/*",
                                                                  remote_name="all")
        except Exception as e:
            Logger().warning(str(e))

        res_list: List[ConanRef] = search_results
        self.info_cache.update_remote_package_list(res_list)
        return res_list

    def get_remote_pkgs_from_ref(self, conan_ref: ConanRef, remote: Optional[str], 
                                 query=None) -> List[ConanPkg]:
        found_pkgs: List[ConanPkg] = []
        try:
            from conan.api.model import ListPattern
            pattern = ListPattern(str(conan_ref) + ":*", rrev="", prev="")
            search_results = None
            remote_obj = None
            if remote:
                for remote_obj in self.get_remotes():
                    if remote_obj.name == remote:
                        break
            search_results = self._conan.list.select(pattern, remote=remote_obj, 
                                                     package_query=query)
            if search_results:
                latest_rev = self._conan.list.latest_recipe_revision(conan_ref, remote_obj)
                if latest_rev:
                    found_pkgs_dict = search_results.recipes.get(str(conan_ref), {}).get(
                        "revisions", {}).get(latest_rev.revision, {}).get("packages", {})
                    for id, info in found_pkgs_dict.items():
                        found_pkgs.append(ConanPkg(id=id, options=info.get("info", {}).get("options", {}),
                                                   settings=info.get("info", {}).get(
                                                       "settings", {}),
                                                   requires=[], outdated=False))
            Logger().debug(str(found_pkgs))
        except ConanException:  # no problem, next
            return []
        return found_pkgs



from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import QEasingCurve, QPoint, QPropertyAnimation, Qt, Slot
from PySide6.QtGui import QAction, QShortcut
from PySide6.QtWidgets import QListWidgetItem, QMenu, QWidget
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer.ui.common.syntax_highlighting import ConfigHighlighter
from conan_explorer.ui.plugin import PluginDescription, PluginInterfaceV1

from .controller import ConanSearchController

if TYPE_CHECKING:
    from conan_explorer.ui.fluent_window import FluentWindow
    from conan_explorer.ui.main_window import BaseSignals


class ConanSearchView(PluginInterfaceV1):
    def __init__(
        self,
        parent: QWidget,
        plugin_description: PluginDescription,
        base_signals: Optional["BaseSignals"],
        page_widgets: Optional["FluentWindow.PageStore"] = None,
    ):
        # Add minimize and maximize buttons
        super().__init__(parent, plugin_description, base_signals, page_widgets)
        from .conan_search_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.load_signal.connect(self.load)

    def load(self):
        conan_pkg_installed = None
        conan_pkg_removed = None
        if self._base_signals:
            conan_pkg_installed = self._base_signals.conan_pkg_installed
            conan_pkg_removed = self._base_signals.conan_pkg_removed

        self._search_controller = ConanSearchController(
            self._ui.search_results_tree_view,
            self._ui.search_line,
            self._ui.search_button,
            self._ui.remote_list,
            self._ui.package_info_text,
            conan_pkg_installed,
            conan_pkg_removed,
        )

        self._ui.search_button.clicked.connect(self._search_controller.on_search)
        self._ui.search_button.setEnabled(False)
        self._ui.install_button.clicked.connect(self._search_controller.on_install_button)
        self._ui.search_line.validator_enabled = False
        self._ui.search_line.textChanged.connect(self._enable_search_button)
        for key in (
            "Enter",
            "Return",
        ):
            self._search_shorcut = QShortcut(
                key, self._ui.search_line, self._ui.search_button.animateClick
            )
            self._search_shorcut.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)

        # init remotes list
        self._init_remotes()
        if self._base_signals:
            self._base_signals.conan_remotes_updated.connect(self._init_remotes)
        self._ui.select_all_widget.itemChanged.connect(self.on_toggled_remotes_selection)

        self._ui.search_results_tree_view.setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._ui.search_results_tree_view.customContextMenuRequested.connect(
            self.on_pkg_context_menu_requested
        )
        self._init_pkg_context_menu()
        # force_light_mode for disabled icon in dark mode
        self.set_themed_icon(
            self._ui.search_button, "icons/search.svg", size=(20, 20), force_light_mode=True
        )
        self.set_themed_icon(self._ui.install_button, "icons/download_pkg.svg", size=(20, 20))
        self._conan_config_highlighter = ConfigHighlighter(
            self._ui.package_info_text.document(), "yaml"
        )
        self._ui.frame.setMinimumHeight(0)
        self._ui.frame.setMaximumHeight(0)
        self.remote_toggle_animation = QPropertyAnimation(self._ui.frame, b"maximumHeight")

        # animation for expanding/collapsing remote list
        def start_animation(checked):
            arrow_type = Qt.ArrowType.DownArrow if checked else Qt.ArrowType.RightArrow
            max_height = self._ui.frame.sizeHint().height()
            start_height = 0 if checked else max_height
            end_height = max_height if checked else 0
            self._ui.remote_toggle_button.setArrowType(arrow_type)
            # self.toggleAnimation.setDirection(direction)
            self.remote_toggle_animation.setDuration(400)
            self.remote_toggle_animation.setStartValue(start_height)
            self.remote_toggle_animation.setEndValue(end_height)
            self.remote_toggle_animation.setEasingCurve(QEasingCurve.Type.InOutQuart)
            self.remote_toggle_animation.start()

        self._ui.remote_toggle_button.clicked.connect(start_animation)

    def _init_remotes(self):
        remotes = app.conan_api.get_remotes()
        self._ui.remote_list.clear()
        for remote in remotes:
            item = QListWidgetItem(remote.name, self._ui.remote_list)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked)

    def on_toggled_remotes_selection(self):
        for i in range(self._ui.remote_list.count()):
            item = self._ui.remote_list.item(i)
            item.setCheckState(self._ui.select_all_widget.item(0).checkState())

    def _enable_search_button(self):
        """Enable search button from minimum 3 characters onwards"""
        if len(self._ui.search_line.text()) > 2:
            self._ui.search_button.setEnabled(True)
            self.set_themed_icon(self._ui.search_button, "icons/search.svg", size=(20, 20))
        else:
            self._ui.search_button.setEnabled(False)
            self.set_themed_icon(
                self._ui.search_button, "icons/search.svg", size=(20, 20), force_light_mode=True
            )

    def _init_pkg_context_menu(self):
        """Initalize context menu with all actions"""
        self.select_cntx_menu = QMenu()
        self.copy_ref_action = QAction("Copy reference", self)
        self.set_themed_icon(self.copy_ref_action, "icons/copy_link.svg")
        self.select_cntx_menu.addAction(self.copy_ref_action)
        self.copy_ref_action.triggered.connect(self._search_controller.on_copy_ref_requested)

        self.show_conanfile_action = QAction("Show conanfile", self)
        self.set_themed_icon(self.show_conanfile_action, "icons/file_preview.svg")
        self.select_cntx_menu.addAction(self.show_conanfile_action)
        self.show_conanfile_action.triggered.connect(
            self._search_controller.on_show_conanfile_requested
        )

        self.install_pkg_action = QAction("Install package", self)
        self.set_themed_icon(self.install_pkg_action, "icons/download_pkg.svg")
        self.select_cntx_menu.addAction(self.install_pkg_action)
        self.install_pkg_action.triggered.connect(
            self._search_controller.on_install_pkg_requested
        )

        self.show_in_pkg_exp_action = QAction("Show in Package Explorer", self)
        self.set_themed_icon(self.show_in_pkg_exp_action, "icons/global/search_packages.svg")
        self.select_cntx_menu.addAction(self.show_in_pkg_exp_action)
        self.show_in_pkg_exp_action.triggered.connect(self.on_show_in_pkg_exp)

        self.diff_pkgs_action = QAction("Compare packages", self)
        self.set_themed_icon(self.diff_pkgs_action, "icons/difference.svg")
        self.select_cntx_menu.addAction(self.diff_pkgs_action)
        self.diff_pkgs_action.triggered.connect(self._search_controller.on_diff_requested)

    def reload_themed_icons(self):
        self._conan_config_highlighter = ConfigHighlighter(
            self._ui.package_info_text.document(), "yaml"
        )
        super().reload_themed_icons()

    @Slot(QPoint)
    def on_pkg_context_menu_requested(self, position: QPoint):
        """
        Executes, when context menu is requested.
        This is done to dynamically grey out some options depending on the item type.
        """
        items = self._search_controller.get_selected_source_items()
        if len(items) < 1:
            return
        elif len(items) < 2:
            self.diff_pkgs_action.setEnabled(False)
        else:
            self.diff_pkgs_action.setEnabled(True)

        item = items[0]
        if item.empty:
            return
        if item.is_installed:
            self.show_in_pkg_exp_action.setEnabled(True)
        else:
            self.show_in_pkg_exp_action.setEnabled(False)
        self.select_cntx_menu.exec(self._ui.search_results_tree_view.mapToGlobal(position))

    def on_show_in_pkg_exp(self):
        """Switch to the Package Explorer view and select the item (ref or pkg)
        Needs other view as ref, so will not be moved to controller.
        """
        from conan_explorer.ui.views import LocalConanPackageExplorer

        items = self._search_controller.get_selected_source_items()
        if len(items) != 1:
            return
        item = items[0]
        if not self._page_widgets:
            return
        self._page_widgets.get_page_by_type(
            LocalConanPackageExplorer
        ).select_local_package_from_ref(item.get_conan_ref())

    @override
    def resizeEvent(self, a0) -> None:
        super().resizeEvent(a0)
        self._search_controller._resize_search_result_columns()

    @override
    def showEvent(self, a0) -> None:
        # set cursor in search line when switching to this tab
        self._ui.search_line.setFocus()
        return super().showEvent(a0)



from typing import Optional

from PySide6.QtWidgets import QDialog, QStyle, QWidget


class QuestionWithItemListDialog(QDialog):
    def __init__(
        self,
        parent: Optional[QWidget],
        icon: QStyle.StandardPixmap = QStyle.StandardPixmap.SP_MessageBoxQuestion,
    ) -> None:
        super().__init__(parent)
        from .qwil_ui import Ui_Dialog

        self._ui = Ui_Dialog()
        self._ui.setupUi(self)
        std_icon = self.style().standardIcon(icon)
        self._ui.icon.setPixmap(std_icon.pixmap(40, 40))
        self.button_box = self._ui.button_box
        self.item_list_widget = self._ui.list_widget

    def set_question_text(self, text: str):
        self._ui.question_label.setText(text)



from .file import PluginFile
from .handler import PluginHandler
from .types import PluginDescription, PluginInterfaceV1

__all__ = ["PluginFile", "PluginHandler", "PluginDescription", "PluginInterfaceV1"]



from pathlib import Path
from typing import Any, Union

from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Qt
from PySide6.QtGui import QFont
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer.settings import PLUGINS_SECTION_NAME
from conan_explorer.ui.common import TreeModel, TreeModelItem, get_themed_asset_icon
from conan_explorer.ui.plugin import PluginDescription, PluginFile, PluginHandler


class PluginModelItem(TreeModelItem):
    def __init__(
        self, plugin: PluginDescription, plugin_path: str, enabled: bool, parent: TreeModelItem
    ):
        super().__init__(
            [
                plugin.name,
                plugin.version,
                plugin.description,
                plugin.conan_versions,
                plugin.author,
            ],
            parent,
            lazy_loading=False,
        )
        self._icon = plugin.icon
        self.enabled = enabled
        self.plugin_path = str(Path(plugin_path).resolve())


class PluginModel(TreeModel):
    def __init__(self, *args, **kwargs):
        super().__init__(checkable=True, *args, **kwargs)
        self.root_item = TreeModelItem(["Name", "Version", "Description", "Conan", "Author"])

    def setup_model_data(self):
        self.root_item.child_items = []
        for plugin_group_name in app.active_settings.get_settings_from_node(
            PLUGINS_SECTION_NAME
        ):
            plugin_path = app.active_settings.get_string(plugin_group_name)
            plugins = PluginFile.read(plugin_path)
            for plugin in plugins:
                plugin_element = PluginModelItem(
                    plugin, plugin_path, PluginHandler.is_plugin_enabled(plugin), self.root_item
                )
                self.root_item.append_child(plugin_element)

    @override
    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: int) -> Any:
        if not index.isValid():
            return None
        item: PluginModelItem = index.internalPointer()  # type: ignore
        if role == Qt.ItemDataRole.DisplayRole:
            try:
                return item.data(index.column())
            except Exception:
                return ""
        if role == Qt.ItemDataRole.DecorationRole:
            if index.column() != 0:  # only display icon for first column
                return
            if isinstance(item, PluginModelItem):
                return get_themed_asset_icon(item._icon)
        if role == Qt.ItemDataRole.FontRole and not item.enabled:
            font = QFont()
            font.setItalic(True)
            return font
        return None



from typing import Union

from PySide6.QtCore import QItemSelectionModel, QModelIndex
from PySide6.QtWidgets import QTreeView

import conan_explorer.app as app
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import EditablePkg

from .editable_model import EditableModel, EditableModelItem


class ConanEditableController:
    def __init__(self, view: QTreeView) -> None:
        self._view = view
        self._model = EditableModel()
        self._view.setModel(self._model)
        self.update()

    def update(self):
        # save selected remote, if triggering a re-init
        sel_edit = self.get_selected_editable()
        self._model.setup_model_data()
        self._view.setItemsExpandable(False)
        self._view.setRootIsDecorated(False)
        self._view.expandAll()
        self.resize_remote_columns()

        if sel_edit:
            self._select_editable(sel_edit.conan_ref)

    def resize_remote_columns(self):
        for i in reversed(range(self._model.root_item.column_count() - 1)):
            self._view.resizeColumnToContents(i)
        # TODO calculate, if we need to make the name smaller
        self._view.setColumnWidth(1, 400)
        self._view.columnViewportPosition(0)

    def _select_editable(self, name: str) -> bool:
        """Selects ane idtable in the view and returns true if it exists."""
        row_remote_to_sel = -1
        row = 0
        editable_item = None
        for editable_item in self._model.root_item.child_items:
            if editable_item.name == name:
                row_remote_to_sel = row
                break
            row += 1
        if row_remote_to_sel < 0:
            Logger().debug("No editable to select")
            return False
        sel_model = self._view.selectionModel()
        sel_model.clearSelection()
        for column in range(self._model.columnCount(QModelIndex())):
            index = self._model.index(row_remote_to_sel, column, QModelIndex())
            sel_model.select(index, QItemSelectionModel.SelectionFlag.Select)
        return True

    def add(self, editable: EditablePkg):
        if not editable:
            return
        return self._model.add(editable)

    def remove(self, editable: EditablePkg):
        if not editable:
            return
        return self._model.remove(editable)

    def get_selected_editable(self) -> Union[EditablePkg, None]:
        indexes = self._view.selectedIndexes()
        if len(indexes) == 0:  # can be multiple - always get 0
            Logger().debug("No selected item for context action")
            return None
        item: EditableModelItem = indexes[0].internalPointer()  # type: ignore
        return app.conan_api.get_editable(item.name)



from typing import Optional

from PySide6.QtCore import QModelIndex, QObject
from PySide6.QtWidgets import QTreeView

from conan_explorer.ui.plugin import PluginHandler

from .model import PluginModel, PluginModelItem


class PluginController(QObject):
    def __init__(self, view: QTreeView, plugin_handler: PluginHandler) -> None:
        super().__init__(view)
        self._view = view
        self._model = PluginModel()
        self._plugin_handler = plugin_handler

    def update(self):
        self._model = PluginModel()
        self._model.setup_model_data()
        self._view.setItemsExpandable(False)
        self._view.setRootIsDecorated(False)
        self._view.setModel(self._model)
        self._view.expandAll()
        self._resize_columns()

    def get_selected_source_item(self) -> Optional[PluginModelItem]:
        """Gets the selected item from a view"""
        indexes = self._view.selectedIndexes()
        if not indexes:
            return None
        view_index = indexes[0]
        source_item: PluginModelItem = view_index.internalPointer()  # type: ignore
        return source_item

    def _resize_columns(self):
        count = self._view.model().columnCount(QModelIndex())
        for i in reversed(range(count - 1)):
            self._view.resizeColumnToContents(i)

    def add_plugin(self, plugin_path: str):
        self._plugin_handler.add_plugin(plugin_path)

    def remove_plugin(self, plugin_path: str):
        self._plugin_handler.remove_plugin(plugin_path)

    def reload_plugin(self, plugin_name: str):
        self._plugin_handler.reload_plugin(plugin_name)



from queue import Queue
from threading import Thread

# this allows to use forward declarations to avoid circular imports
from typing import TYPE_CHECKING, Any, List, Optional, Protocol, Tuple, TypedDict

from conan_explorer.settings import SettingsInterface

if TYPE_CHECKING:
    from ..conan_wrapper import ConanCommonUnifiedApi

from conan_explorer import USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL
from conan_explorer.app.logger import Logger

from .types import ConanOptions, ConanPkgRef, ConanRef, ConanSettings


class ConanWorkerElement(TypedDict):
    # If id is used options, settings and auto_install is ignored
    ref_pkg_id: str  # format in <ref>:<id>. Id is optional.
    options: ConanOptions  # conan options with key-value pairs
    settings: ConanSettings  # conan settings with key-value pairs
    profile: str  # alternative to settings
    update: bool  # use -u flag for install
    auto_install: bool  # automatically determine best matching package.


class ConanWorkerResultCallback(Protocol):
    def __call__(self, conan_ref: str, pkg_id: str) -> Any: ...


class ConanWorker:
    """
    Sequential worker with a queue to execute conan install/version alternatives commands
    """

    def __init__(self, conan_api: "ConanCommonUnifiedApi", settings: SettingsInterface):
        self._conan_api = conan_api
        self._conan_install_queue: Queue[
            Tuple[ConanWorkerElement, Optional[ConanWorkerResultCallback]]
        ] = Queue(maxsize=0)
        self._install_worker: Optional[Thread] = None
        self._shutdown_requested = False  # internal flag to cancel worker on shutdown
        self._settings = settings

    def update_all_info(
        self,
        conan_elements: List[ConanWorkerElement],
        info_callback: Optional[ConanWorkerResultCallback],
    ):
        """
        Starts the worker for all given elements. Should be called at start.
        info_signal is used to notify the caller that the worker has finished
        for a given element. To be able to identify, which package has been finished
        the signal will send this info: tuple(conan_ref, pkg_id)
        """
        # fill up queue
        for worker_element in conan_elements:
            if USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL:
                self._conan_install_queue.put((worker_element, info_callback))

        # start getting versions info in a separate thread in a bundled way
        # to get better performance
        self._start_install_worker()

    def put_ref_in_install_queue(
        self,
        conan_element: ConanWorkerElement,
        info_callback: Optional[ConanWorkerResultCallback],
    ):
        """Add a new entry to work on"""
        self._conan_install_queue.put((conan_element, info_callback))
        self._start_install_worker()

    def _start_install_worker(self):
        """Start worker, if it is not already started (can be called multiple times)"""
        if not self._install_worker or not self._install_worker.is_alive():
            self._install_worker = Thread(
                target=self._work_on_conan_install_queue, name="ConanInstallWorker", daemon=True
            )
            self._install_worker.start()

    def _work_on_conan_install_queue(self):
        """Call conan install from queue"""
        info_callback = None
        pkg_id = ""
        while not self._shutdown_requested and not self._conan_install_queue.empty():
            worker_element, info_callback = self._conan_install_queue.get()
            ref_pkg_id = worker_element.get("ref_pkg_id", "")
            conan_options = worker_element.get("options", {})
            conan_settings = worker_element.get("settings", {})
            conan_profile = worker_element.get("profile", "")
            pkg_id = ""
            update = worker_element.get("update", False)
            auto_install = worker_element.get("auto_install", True)
            # package path will be updated in conan cache
            try:
                conan_ref: ConanRef
                if ":" in ref_pkg_id:  # pkg ref
                    pkg_ref: ConanPkgRef = ConanPkgRef.loads(ref_pkg_id)  # type: ignore
                    conan_ref = pkg_ref.ref  # type: ignore
                    package = self._conan_api.get_remote_pkg_from_id(pkg_ref)
                    pkg_id, _ = self._conan_api.install_package(pkg_ref.ref, package, update)
                else:
                    conan_ref = ConanRef.loads(ref_pkg_id)  # type: ignore

                    if auto_install:
                        pkg_id, _ = self._conan_api.get_path_or_auto_install(
                            conan_ref, conan_options, update
                        )
                    else:
                        pkg_id, _ = self._conan_api.install_reference(
                            conan_ref, conan_settings, conan_options, conan_profile, update
                        )
            except Exception:
                try:
                    self._conan_install_queue.task_done()
                except ValueError:
                    pass  # don't care about calling too many times
                continue
            Logger().debug("Finish working on " + ref_pkg_id)
            try:
                self._conan_install_queue.task_done()
            except ValueError:
                pass  # don't care about calling too many times
            if USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL:
                if info_callback and not self._shutdown_requested:
                    try:
                        info_callback(self._conan_api.generate_canonical_ref(conan_ref), pkg_id)
                    except Exception as e:
                        Logger().error(str(e))

    def finish_working(self, timeout_s: Optional[float] = None):
        """Cancel, if worker is still not finished"""
        self._shutdown_requested = True
        try:
            if self._install_worker and self._install_worker.is_alive():
                self._install_worker.join(timeout_s)
        except Exception:
            return  # Conan threads can crash on join
        self._conan_install_queue = Queue(maxsize=0)
        self._install_worker = None  # reset thread for later instantiation
        self._shutdown_requested = False



from pathlib import Path

from jinja2 import Template
from PySide6.QtGui import QColor, QFont, QFontDatabase, QPalette
from PySide6.QtWidgets import QApplication

import conan_explorer.app as app
from conan_explorer import base_path
from conan_explorer.app.logger import Logger
from conan_explorer.app.system import is_windows_11
from conan_explorer.settings import FONT_SIZE, GUI_MODE, GUI_MODE_DARK, GUI_MODE_LIGHT


def activate_theme(qt_app: QApplication):
    """Apply the theme from the current settings and apply all related view options"""
    dark_mode = get_gui_dark_mode()
    style_file = "light_style.qss.in"
    if dark_mode:
        style_file = "dark_style.qss.in"
    try:
        user_color = QPalette().accent().color().name()
    except Exception:
        user_color = "#000000"
    window_border_radius = 0

    # enable rounded corners under Win 11
    if is_windows_11():
        window_border_radius = 7

    # apply palette for correct icon theme detection under new windows platform plugin
    if dark_mode:
        qt_app.setPalette(QPalette(QColor("#000000")))
    else:
        qt_app.setPalette(QPalette(QColor("#FFFFFF")))

    style_sheet = configure_theme(
        base_path / "ui" / style_file,
        app.active_settings.get_int(FONT_SIZE),
        user_color,
        window_border_radius,
    )

    qt_app.setStyleSheet(style_sheet)


def configure_theme(
    qss_template_path: Path, font_size_pt: int, user_color: str, window_border_radius: int
) -> str:
    """Configure the given qss file with the set options and return it as a string"""

    qss_template = None
    register_font("Noto Sans Mono", "NotoSansMono.ttf")
    register_font("Noto Sans", "NotoSans-Regular.ttf")
    with open(qss_template_path, "r") as fd:
        qss_template = Template(fd.read())
    qss_content = qss_template.render(
        MAIN_FONT_SIZE=font_size_pt,
        USER_COLOR=user_color,
        WINDOW_BORDER_RADIUS=window_border_radius,
        CONSOLE_FONT_FAMILY="Noto Sans Mono",
        FONT_FAMILY="Noto Sans",
    )
    return qss_content


def get_gui_dark_mode() -> bool:
    gui_mode = app.active_settings.get_string(GUI_MODE).lower()
    if gui_mode not in [GUI_MODE_DARK, GUI_MODE_LIGHT]:
        Logger().warning("Theming: Can't read GUI mode, setting it to light.")
        return False
    return True if gui_mode == GUI_MODE_DARK else False


def register_font(font_style_name: str, font_file_name: str) -> "QFont":
    # set up font
    font_file = app.asset_path / "font" / font_file_name
    font_id = QFontDatabase.addApplicationFont(str(font_file))
    qapp = QApplication.instance()
    if qapp is None:
        return QFont()
    font = qapp.font()  # type: ignore
    if font_id != -1:
        font_db = QFontDatabase()
        font_styles = font_db.styles(font_style_name)
        font_families = QFontDatabase.applicationFontFamilies(font_id)
        if font_families:
            font = font_db.font(font_families[0], font_styles[0], 13)
            font.setStyleStrategy(QFont.StyleStrategy.PreferAntialias)
        else:
            Logger().warning("Can't register selected font file.")
    return font



"""
On catching an uncaught execution show a dialog with a link to post an issue,
with prefilled text and stacktrace.
"""

import platform
import traceback
from types import TracebackType
from typing import Optional

from PySide6.QtWidgets import QStyle

from conan_explorer import REPO_URL, __version__

bug_dialog_text = f"""
**Describe the bug**
Application crash.

**To Reproduce**
Steps to reproduce the behavior:
1. Go to '...'
2. Click on '....'
3. Scroll down to '....'
4. See error

**Screenshots**
If applicable, add screenshots to help explain your problem.

**Desktop:**
 - OS: {platform.platform(terse=True)}
 - Python Version: {platform.python_version()}
 - Version: {__version__}

**Additional context**
Add any other context about the problem here.
"""


def show_bug_reporting_dialog(excvalue: Optional[BaseException], tb: "TracebackType | None"):
    import urllib.parse  # late import hopefully we don't need this

    error_text = f"{str(excvalue)}\n" + "\n".join(traceback.format_tb(tb, limit=None))
    title = urllib.parse.quote("Application Crash on <>")
    body = urllib.parse.quote(f"{bug_dialog_text}\n**Stacktrace**:\n" + error_text)
    new_issue_with_info_text = f"{REPO_URL}/issues/new?title={title}&body={body}&labels=bug"
    html_crash_text = f'\
        <html><head/><body><p> \
        Oops, something went wrong!\
        To help improve the program, please post a <a href="{new_issue_with_info_text}"> \
        <span style=" text-decoration: underline color:  # 0000ff;" \
        >new github issue</span></a> and describe, how the crash occured!'

    from .crash_ui import QDialog, Ui_Dialog

    dialog = QDialog()
    dialog_ui = Ui_Dialog()
    dialog_ui.setupUi(dialog)
    dialog_ui.crash_message_label.setText(html_crash_text)
    dialog_ui.error_text_browser.setText(error_text)
    pixmapi = QStyle.StandardPixmap.SP_MessageBoxCritical
    icon = dialog.style().standardIcon(pixmapi)
    dialog.setWindowIcon(icon)
    dialog.setMinimumWidth(800)
    dialog.exec()
    return dialog  # for testing



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'app_edit_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QGridLayout, QLabel, QLayout, QLineEdit,
    QPushButton, QSizePolicy, QTextEdit, QToolButton,
    QVBoxLayout, QWidget)

from conan_explorer.ui.widgets import AnimatedToggle
from conan_explorer.ui.widgets.conan_line_edit import ConanRefLineEdit

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(769, 568)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setModal(True)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.edit_grid = QGridLayout()
        self.edit_grid.setObjectName(u"edit_grid")
        self.edit_grid.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.edit_grid.setVerticalSpacing(8)
        self.conan_ref_line_edit = ConanRefLineEdit(Dialog)
        self.conan_ref_line_edit.setObjectName(u"conan_ref_line_edit")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.conan_ref_line_edit.sizePolicy().hasHeightForWidth())
        self.conan_ref_line_edit.setSizePolicy(sizePolicy1)
        self.conan_ref_line_edit.setMinimumSize(QSize(400, 0))

        self.edit_grid.addWidget(self.conan_ref_line_edit, 1, 1, 1, 1)

        self.icon_line_edit = QLineEdit(Dialog)
        self.icon_line_edit.setObjectName(u"icon_line_edit")

        self.edit_grid.addWidget(self.icon_line_edit, 6, 1, 1, 1)

        self.args_line_edit = QLineEdit(Dialog)
        self.args_line_edit.setObjectName(u"args_line_edit")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.args_line_edit.sizePolicy().hasHeightForWidth())
        self.args_line_edit.setSizePolicy(sizePolicy2)

        self.edit_grid.addWidget(self.args_line_edit, 4, 1, 1, 1)

        self.execpath_line_edit = QLineEdit(Dialog)
        self.execpath_line_edit.setObjectName(u"execpath_line_edit")

        self.edit_grid.addWidget(self.execpath_line_edit, 3, 1, 1, 1)

        self.execpath_label = QLabel(Dialog)
        self.execpath_label.setObjectName(u"execpath_label")
        self.execpath_label.setWordWrap(True)

        self.edit_grid.addWidget(self.execpath_label, 3, 0, 1, 1)

        self.conan_opts_text_edit = QTextEdit(Dialog)
        self.conan_opts_text_edit.setObjectName(u"conan_opts_text_edit")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.conan_opts_text_edit.sizePolicy().hasHeightForWidth())
        self.conan_opts_text_edit.setSizePolicy(sizePolicy3)
        self.conan_opts_text_edit.setMinimumSize(QSize(600, 0))
        self.conan_opts_text_edit.setAutoFormatting(QTextEdit.AutoNone)

        self.edit_grid.addWidget(self.conan_opts_text_edit, 2, 1, 1, 1)

        self.name_line_edit = QLineEdit(Dialog)
        self.name_line_edit.setObjectName(u"name_line_edit")
        self.name_line_edit.setMinimumSize(QSize(500, 0))

        self.edit_grid.addWidget(self.name_line_edit, 0, 1, 1, 1)

        self.icon_label = QLabel(Dialog)
        self.icon_label.setObjectName(u"icon_label")
        self.icon_label.setTextFormat(Qt.RichText)
        self.icon_label.setWordWrap(True)

        self.edit_grid.addWidget(self.icon_label, 6, 0, 1, 1)

        self.icon_browse_button = QToolButton(Dialog)
        self.icon_browse_button.setObjectName(u"icon_browse_button")
        sizePolicy2.setHeightForWidth(self.icon_browse_button.sizePolicy().hasHeightForWidth())
        self.icon_browse_button.setSizePolicy(sizePolicy2)

        self.edit_grid.addWidget(self.icon_browse_button, 6, 2, 1, 1)

        self.conan_opts_label = QLabel(Dialog)
        self.conan_opts_label.setObjectName(u"conan_opts_label")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.conan_opts_label.sizePolicy().hasHeightForWidth())
        self.conan_opts_label.setSizePolicy(sizePolicy4)

        self.edit_grid.addWidget(self.conan_opts_label, 2, 0, 1, 1)

        self.args_label = QLabel(Dialog)
        self.args_label.setObjectName(u"args_label")

        self.edit_grid.addWidget(self.args_label, 4, 0, 1, 1)

        self.is_console_app_label = QLabel(Dialog)
        self.is_console_app_label.setObjectName(u"is_console_app_label")

        self.edit_grid.addWidget(self.is_console_app_label, 5, 0, 1, 1)

        self.executable_browse_button = QToolButton(Dialog)
        self.executable_browse_button.setObjectName(u"executable_browse_button")
        sizePolicy2.setHeightForWidth(self.executable_browse_button.sizePolicy().hasHeightForWidth())
        self.executable_browse_button.setSizePolicy(sizePolicy2)
        self.executable_browse_button.setLayoutDirection(Qt.LeftToRight)

        self.edit_grid.addWidget(self.executable_browse_button, 3, 2, 1, 1)

        self.conan_ref_label = QLabel(Dialog)
        self.conan_ref_label.setObjectName(u"conan_ref_label")
        sizePolicy3.setHeightForWidth(self.conan_ref_label.sizePolicy().hasHeightForWidth())
        self.conan_ref_label.setSizePolicy(sizePolicy3)
        self.conan_ref_label.setMinimumSize(QSize(0, 40))
        self.conan_ref_label.setWordWrap(True)

        self.edit_grid.addWidget(self.conan_ref_label, 1, 0, 1, 1)

        self.is_console_app_checkbox = AnimatedToggle(Dialog)
        self.is_console_app_checkbox.setObjectName(u"is_console_app_checkbox")

        self.edit_grid.addWidget(self.is_console_app_checkbox, 5, 1, 1, 1)

        self.name_label = QLabel(Dialog)
        self.name_label.setObjectName(u"name_label")
        self.name_label.setWordWrap(True)

        self.edit_grid.addWidget(self.name_label, 0, 0, 1, 1)

        self.install_button = QPushButton(Dialog)
        self.install_button.setObjectName(u"install_button")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.install_button.sizePolicy().hasHeightForWidth())
        self.install_button.setSizePolicy(sizePolicy5)
        self.install_button.setFlat(False)

        self.edit_grid.addWidget(self.install_button, 7, 1, 1, 1)

        self.edit_grid.setRowStretch(0, 1)
        self.edit_grid.setRowStretch(1, 1)
        self.edit_grid.setColumnStretch(1, 1)

        self.verticalLayout.addLayout(self.edit_grid)

        self.note_label = QLabel(Dialog)
        self.note_label.setObjectName(u"note_label")

        self.verticalLayout.addWidget(self.note_label)

        self.button_box = QDialogButtonBox(Dialog)
        self.button_box.setObjectName(u"button_box")
        sizePolicy6 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy6.setHorizontalStretch(1)
        sizePolicy6.setVerticalStretch(1)
        sizePolicy6.setHeightForWidth(self.button_box.sizePolicy().hasHeightForWidth())
        self.button_box.setSizePolicy(sizePolicy6)
        self.button_box.setMinimumSize(QSize(0, 0))
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(Dialog)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Edit App Link", None))
        self.args_line_edit.setText("")
        self.execpath_label.setText(QCoreApplication.translate("Dialog", u"Relative path of file to open", None))
        self.conan_opts_text_edit.setHtml(QCoreApplication.translate("Dialog", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'Segoe UI'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'MS Shell Dlg 2'; font-size:7.875pt;\"><br /></p></body></html>", None))
        self.icon_label.setText(QCoreApplication.translate("Dialog", u"Icon path (relative or absoulte)", None))
        self.icon_browse_button.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.conan_opts_label.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Conan Options</p><p>using format: </p><p>name1=val1</p><p>name2=val2</p></body></html>", None))
        self.args_label.setText(QCoreApplication.translate("Dialog", u"Launch arguments", None))
        self.is_console_app_label.setText(QCoreApplication.translate("Dialog", u"Start in new console", None))
        self.executable_browse_button.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.conan_ref_label.setText(QCoreApplication.translate("Dialog", u"Conan Reference", None))
        self.is_console_app_checkbox.setText("")
        self.name_label.setText(QCoreApplication.translate("Dialog", u"Link Name", None))
        self.install_button.setText(QCoreApplication.translate("Dialog", u"Install package", None))
        self.note_label.setText(QCoreApplication.translate("Dialog", u"All relative paths resolve to the Conan package!", None))
    # retranslateUi




from pathlib import Path
from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QDialog, QFrame, QMenu, QMessageBox, QWidget
from typing_extensions import override

from conan_explorer import ICON_SIZE, INVALID_PATH
from conan_explorer.app.logger import Logger
from conan_explorer.app.system import open_in_file_manager, run_file
from conan_explorer.ui.common import get_themed_asset_icon, measure_font_width
from conan_explorer.ui.dialogs.reorder import ReorderDialog

from .dialogs import AppEditDialog
from .model import UiAppLinkModel

if TYPE_CHECKING:
    from .tab import TabList

OFFICIAL_RELEASE_DISP_NAME = "<official release>"
OFFICIAL_USER_DISP_NAME = "<official user>"

current_dir = Path(__file__).parent


class ListAppLink(QFrame):
    """Represents a clickable button + info for an executable in a conan package.
    Rightclick context menu has the following elements:
    - Show in File Manager
    - Add new App Link
    - Edit
    - Remove App Link
    - Reorder App Links
    """

    icon_size: int

    def __init__(
        self,
        parent: Optional[QWidget],
        parent_tab: "TabList",
        model: UiAppLinkModel,
        icon_size=ICON_SIZE,
    ):
        super().__init__(parent)
        self.setObjectName(repr(self))
        self.icon_size = icon_size
        self.model = model
        self._parent_tab = parent_tab  # save parent - don't use qt signals ands slots

        from .app_link_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)

        self._ui.edit_button.setIcon(QIcon(get_themed_asset_icon("icons/edit.svg")))
        self._ui.remove_button.setIcon(QIcon(get_themed_asset_icon("icons/delete.svg")))
        self._ui.app_button.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._ui.app_button.customContextMenuRequested.connect(self.on_context_menu_requested)

        # connect signals
        self._ui.app_button.clicked.connect(self.on_click)
        self._ui.edit_button.clicked.connect(self.open_edit_dialog)
        self._ui.remove_button.clicked.connect(self.remove)
        self._init_context_menu()

    def _init_context_menu(self):
        """Setup context menu."""
        self.menu = QMenu()
        self.open_fm_action = QAction("Show in File Manager", self)
        self.open_fm_action.setIcon(QIcon(get_themed_asset_icon("icons/file_explorer.svg")))
        self.menu.addAction(self.open_fm_action)
        self.open_fm_action.triggered.connect(self.on_open_in_file_manager)

        self.menu.addSeparator()

        self.add_action = QAction("Add new App Link", self)
        self.add_action.setIcon(QIcon(get_themed_asset_icon("icons/add_link.svg")))
        self.menu.addAction(self.add_action)
        self.add_action.triggered.connect(self.open_app_link_add_dialog)

        self.edit_action = QAction("Edit", self)
        self.edit_action.setIcon(QIcon(get_themed_asset_icon("icons/edit.svg")))
        self.menu.addAction(self.edit_action)
        self.edit_action.triggered.connect(self.open_edit_dialog)

        self.remove_action = QAction("Remove App Link", self)
        self.remove_action.setIcon(QIcon(get_themed_asset_icon("icons/delete.svg")))
        self.menu.addAction(self.remove_action)
        self.remove_action.triggered.connect(self.remove)

        self.menu.addSeparator()

        self.reorder_action = QAction("Reorder App Links", self)
        self.reorder_action.setIcon(QIcon(get_themed_asset_icon("icons/rearrange.svg")))
        self.reorder_action.triggered.connect(self.on_move)

        self.menu.addAction(self.reorder_action)

    def load(self):
        self.model.register_update_callback(self.apply_conan_info)
        self._apply_new_config()

    def on_move(self):
        move_dialog = ReorderDialog(parent=self, model=self.model.parent)
        ret = move_dialog.exec()
        if ret == QDialog.DialogCode.Accepted:
            self._parent_tab.redraw(force=True)

    @override
    def resizeEvent(self, event):
        if not self._parent_tab:
            return
        content_frame: QWidget = self._parent_tab.parent().parent().parent().parent().parent()  # type: ignore
        max_cl_width = (
            content_frame.width() - self._ui.left_frame.width() - self._ui.right_frame.width()
        )
        if max_cl_width < 400:  # TODO find better solution
            self._ui.central_left_frame.setMaximumWidth(0)
            self._ui.central_right_frame.setMaximumWidth(0)
            self._ui.arguments_name_label.setMaximumWidth(0)
            self._ui.arguments_value_label.setMaximumWidth(0)
            self._ui.arguments_value_label.setText("")
            return
        else:
            self._ui.central_left_frame.setMaximumWidth(10000)

        self._ui.central_left_frame.adjustSize()
        max_sum_width = (
            content_frame.width()
            - self._ui.left_frame.width()
            - self._ui.central_left_frame.width()
            - self._ui.right_frame.width()
        )

        # Hide arguments, if too big
        if max_sum_width < 250:
            self._ui.central_right_frame.setMaximumWidth(0)
            self._ui.arguments_name_label.setMaximumWidth(0)
            self._ui.arguments_value_label.setMaximumWidth(0)
            self._ui.arguments_value_label.setText("")

        else:
            self._ui.central_right_frame.setMaximumWidth(max_sum_width)
            self._ui.arguments_name_label.setMaximumWidth(1000)
            self._ui.arguments_name_label.adjustSize()
            self._ui.arguments_value_label.setMaximumWidth(
                max_sum_width - self._ui.arguments_name_label.width() - 50
            )
            self.split_into_lines(
                self._ui.arguments_value_label,
                self.model.args,
                max_sum_width - self._ui.arguments_name_label.width(),
            )

        super().resizeEvent(event)

    def split_into_lines(self, widget, model_value, max_width):
        """Calculate, how text can be split into multiple lines, based on the current width"""
        px = measure_font_width(model_value)
        if px == 0:
            return
        new_length = int(len(model_value) * (max_width - 10) / px)
        if len(widget.text().split("\n")[0]) > new_length > len(
            model_value
        ) or new_length - 1 == len(widget.text().split("\n")[0]):
            return
        args = self.word_wrap(model_value, new_length)
        widget.setText(args)

    @staticmethod
    def word_wrap(text: str, max_length: int) -> str:
        if max_length == 0:
            return text
        split_name = text.split(" ")
        name = ""  # split long titles
        for word in split_name:
            if len(word) < max_length:
                new_word = word
            else:
                n_to_short = int(len(word) / max_length) + int(len(word) % max_length > 0)
                new_word = ""
                for i in range(n_to_short):
                    new_word += word[max_length * i : max_length * (i + 1)] + "\n"
                new_word = new_word[:-1]  # remove last \n
            name += " " + new_word if name else new_word
        return name

    def on_context_menu_requested(self, position):
        self.menu.exec(self._ui.app_button.mapToGlobal(position))

    def on_open_in_file_manager(self):
        open_in_file_manager(self.model.get_executable_path().parent)

    def _apply_new_config(self):
        self._ui.app_name_label.setText(self.model.name)
        self._ui.conan_ref_value_label.setText(self.model.conan_ref)
        self._ui.app_button.set_icon(self.model.get_icon())
        self._ui.executable_value_label.setText(self.model.executable)
        self._ui.arguments_value_label.setText(self.model.args)
        self._ui.open_shell_checkbox.setChecked(self.model.is_console_application)

        self.apply_conan_info()  # update with offline information

    def open_app_link_add_dialog(self):
        self._parent_tab.open_app_link_add_dialog()

    def open_edit_dialog(self, model: Optional[UiAppLinkModel] = None):
        if model:
            self.model = model
        edit_app_dialog = AppEditDialog(self.model, parent=self)
        reply = edit_app_dialog.exec()
        if reply == AppEditDialog.DialogCode.Accepted:
            # grey icon, so update from cache can ungrey it, if the path is correct
            self._ui.app_button.grey_icon()
            self.model.load_from_cache()
            # now apply gui config with resolved paths
            self._apply_new_config()
        del edit_app_dialog  # call delete manually for faster thread cleanup

    def remove(self):
        # confirmation dialog
        message_box = QMessageBox(parent=self)
        message_box.setWindowTitle("Delete app link")
        message_box.setText(f'Are you sure, you want to delete the link "{self.model.name}"?')
        message_box.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        message_box.setIcon(QMessageBox.Icon.Question)
        reply = message_box.exec()
        if reply == QMessageBox.StandardButton.Yes:
            self.hide()
            self.model.parent.apps.remove(self.model)
            self._parent_tab.app_links.remove(self)

            self.model.save()
            self._parent_tab.redraw(force=True)

    def update_icon(self):
        self._ui.app_button.set_icon(self.model.get_icon())
        if self.model.get_executable_path() != Path(INVALID_PATH):
            self._ui.app_button.ungrey_icon()

    def on_click(self):
        """Callback for opening the executable on click"""
        if not self.model.get_executable_path().is_file():
            Logger().error(
                (
                    "Can't find file in package "
                    f"{self.model.conan_ref}:\n    {str(self.model._executable)}"
                )
            )
        run_file(
            self.model.get_executable_path(), self.model.is_console_application, self.model.args
        )

    def apply_conan_info(self):
        """Update with new conan data"""
        self.update_icon()



from typing import Dict, List, Optional

from PySide6.QtCore import Qt, SignalInstance
from PySide6.QtWidgets import QDialogButtonBox, QListWidgetItem, QWidget

import conan_explorer.app as app
from conan_explorer.app import LoaderGui  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanRef

from .. import QuestionWithItemListDialog

StdButton = QDialogButtonBox.StandardButton


class ConanRemoveDialog(QuestionWithItemListDialog):
    def __init__(
        self,
        parent: Optional[QWidget],
        conan_refs_with_pkg_ids: Dict[str, List[str]],
        conan_pkg_removed: Optional[SignalInstance] = None,
    ):
        super().__init__(parent)
        self.setWindowTitle("Remove Package(s)")
        self.set_question_text("Are you sure you want to remove these packages?")
        self._conan_pkg_removed_sig = conan_pkg_removed
        for conan_ref, pkg_ids in conan_refs_with_pkg_ids.items():
            for pkg_id in pkg_ids:
                if not pkg_id:
                    text = conan_ref
                else:
                    text = f"{conan_ref}:{pkg_id}"
                list_item = QListWidgetItem(text)
                list_item.setCheckState(Qt.CheckState.Checked)
                self.item_list_widget.addItem(list_item)

        self.button_box.button(StdButton.Yes).clicked.connect(self.on_remove)

    def on_remove(self):
        """Remove conan ref/pkg and emit a signal, if registered"""
        self.loader = LoaderGui(self)
        self.loader.load_for_blocking(
            self, self.remove, cancel_button=False, loading_text="Removing packages"
        )
        self.loader.wait_for_finished()

    def remove(self):
        """Remove the selected ref or pkg. Emit conan_pkg_removed global signal.
        To be called while loading dialog is active."""
        for list_row in range(self.item_list_widget.count()):
            list_item = self.item_list_widget.item(list_row)
            if list_item.checkState() != Qt.CheckState.Checked:
                continue
            conan_ref_with_id = list_item.text()
            pkg_id = ""
            if ":" in conan_ref_with_id:
                conan_ref, pkg_id = conan_ref_with_id.split(":")
            else:
                conan_ref = conan_ref_with_id
            try:
                self.loader.loading_string_signal.emit(f"Removing {conan_ref} {pkg_id}")
                # can handle multiple pkgs at once, but then we can't log info for the
                # progress bar
                app.conan_api.remove_reference(ConanRef.loads(conan_ref), pkg_id)
            except Exception as e:
                Logger().error(f"Error while removing package {conan_ref}: {str(e)}")
                continue
            if self._conan_pkg_removed_sig:
                self._conan_pkg_removed_sig.emit(conan_ref, pkg_id)
            Logger().info(f"Removing package {conan_ref_with_id} finished")




from typing import Optional

from PySide6.QtWidgets import QDialog, QWidget

from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import Remote
from conan_explorer.ui.common.theming import get_themed_asset_icon
from conan_explorer.ui.views.conan_conf.remotes_controller import \
    ConanRemoteController


class RemoteEditDialog(QDialog):

    def __init__(self, remote: Optional[Remote], remotes_controller: ConanRemoteController, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        from .remote_edit_dialog_ui import Ui_Dialog
        self._new_remote = False
        self._remotes_controller = remotes_controller
        if remote is None:
            remote = Remote("New", "", True, False)
            self._new_remote = True
        self._remote = remote

        self._ui = Ui_Dialog()
        self._ui.setupUi(self)

        self.setWindowIcon(get_themed_asset_icon("icons/edit.svg", True))
        self._ui.button_box.accepted.connect(self.save)

        self._ui.name_line_edit.setText(remote.name)
        self._ui.url_line_edit.setText(remote.url)
        self._ui.verify_ssl_checkbox.setChecked(bool(remote.verify_ssl))

    def save(self):
        """ Save edited remote information by calling the appropriate conan methods. """
        new_name = self._ui.name_line_edit.text()
        new_url = self._ui.url_line_edit.text()
        new_verify_ssl = self._ui.verify_ssl_checkbox.isChecked()
        try:
            if self._new_remote:
                self._remotes_controller.add(
                    Remote(new_name, new_url, new_verify_ssl, False))
                self.accept()
                return
            if new_name != self._remote.name:
                self._remotes_controller.rename(self._remote, new_name)
            if new_url != self._remote.url or new_verify_ssl != self._remote.verify_ssl:
                self._remotes_controller.update_remote(
                    Remote(new_name, new_url, new_verify_ssl, self._remote.disabled))
        except Exception as e:
            Logger().error(str(e))
        self.accept()



import platform
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QApplication, QInputDialog, QMenu, QMessageBox, QWidget
from typing_extensions import override

import conan_explorer.app as app
from conan_explorer import conan_version
from conan_explorer.app.logger import Logger
from conan_explorer.app.system import delete_path
from conan_explorer.ui.common import ConfigHighlighter, get_themed_asset_icon
from conan_explorer.ui.plugin import PluginDescription, PluginInterfaceV1

from .dialogs import EditableEditDialog, RemoteEditDialog, RemoteLoginDialog
from .editable_controller import ConanEditableController
from .profiles_model import ProfilesModel
from .remotes_controller import ConanRemoteController

if TYPE_CHECKING:
    from conan_explorer.ui.fluent_window.fluent_window import FluentWindow
    from conan_explorer.ui.main_window import BaseSignals


class ConanConfigView(PluginInterfaceV1):
    load_signal = Signal()  # type: ignore

    def __init__(
        self,
        parent: QWidget,
        plugin_description: PluginDescription,
        base_signals: "BaseSignals",
        page_widgets: Optional["FluentWindow.PageStore"] = None,
    ):
        super().__init__(parent, plugin_description, base_signals)
        from .conan_conf_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.load_signal.connect(self.load)
        self.profiles_path = Path("Unknown")
        self._edited_profile = ""

    def load(self):
        assert self._base_signals
        self._remotes_controller = ConanRemoteController(
            self._ui.remotes_tree_view, self._base_signals.conan_remotes_updated
        )
        self._editable_controller = ConanEditableController(self._ui.editables_ref_view)
        self._init_remotes_tab()
        self._init_profiles_tab()

        self.config_file_path = app.conan_api.get_config_file_path()
        self.profiles_path = app.conan_api.get_profiles_path()
        self._load_info_tab()
        self._load_remotes_tab()
        self._load_profiles_tab()
        self._load_config_file_tab()
        self._load_settings_yml_tab()
        self._load_editables_tab()

        # always show first tab on start
        self._ui.config_tab_widget.tabBar().setCurrentIndex(0)
        self._conan_config_highlighter = ConfigHighlighter(
            self._ui.config_file_text_browser.document(), "ini"
        )
        self._profile_highlighter = ConfigHighlighter(
            self._ui.profiles_text_browser.document(), "ini"
        )
        self._settings_highlighter = ConfigHighlighter(
            self._ui.settings_file_text_browser.document(), "yaml"
        )

    def _load_info_tab(self):
        self._ui.conan_cur_version_value_label.setText(str(conan_version))
        self._ui.python_exe_value_label.setText(str(Path(sys.executable).resolve()))
        self._ui.python_cur_version_value_label.setText(platform.python_version())
        self._ui.revision_enabled_checkbox.setChecked(app.conan_api.get_revisions_enabled())
        self._ui.conan_usr_home_value_label.setText(
            str(app.conan_api.get_user_home_path().resolve())
        )
        if conan_version.major == 2:
            self._ui.conan_usr_cache_value_label.setVisible(False)
            self._ui.conan_usr_cache_label.setVisible(False)
        else:
            self._ui.conan_usr_cache_value_label.setText(
                str(app.conan_api.get_short_path_root().resolve())
            )
        self._ui.conan_storage_path_value_label.setText(
            str(app.conan_api.get_storage_path().resolve())
        )

    def _load_settings_yml_tab(self):
        try:
            self._ui.settings_file_text_browser.setText(
                app.conan_api.get_settings_file_path().read_text()
            )
        except Exception:
            Logger().error("Cannot read settings.yaml file!")

    @override
    def resizeEvent(self, a0):
        """Resize remote view columns automatically if window size changes"""
        super().resizeEvent(a0)
        self._remotes_controller.resize_remote_columns()

    def reload_themed_icons(self):
        super().reload_themed_icons()
        self._init_profile_context_menu()
        self._init_remote_context_menu()
        self._conan_config_highlighter = ConfigHighlighter(
            self._ui.config_file_text_browser.document(), "ini"
        )
        self._profile_highlighter = ConfigHighlighter(
            self._ui.profiles_text_browser.document(), "ini"
        )
        self._settings_highlighter = ConfigHighlighter(
            self._ui.settings_file_text_browser.document(), "yaml"
        )

    # Profile

    def _init_profiles_tab(self):
        self._ui.profiles_list_view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._ui.profiles_list_view.customContextMenuRequested.connect(
            self.on_profile_context_menu_requested
        )
        self._init_profile_context_menu()
        self._ui.profile_save_button.clicked.connect(self.on_profile_save_file)
        self._ui.profile_add_button.clicked.connect(self.on_profile_add)
        self._ui.profile_remove_button.clicked.connect(self.on_profile_remove)
        self._ui.profile_rename_button.clicked.connect(self.on_profile_rename)
        self._ui.profile_refresh_button.clicked.connect(self.on_profiles_refresh)
        self._ui.profiles_copy_name_button.clicked.connect(self.on_copy_profile_requested)

        self.set_themed_icon(self._ui.profile_save_button, "icons/save.svg")
        self.set_themed_icon(self._ui.profile_add_button, "icons/plus_rounded.svg")
        self.set_themed_icon(self._ui.profile_remove_button, "icons/delete.svg")
        self.set_themed_icon(self._ui.profile_refresh_button, "icons/refresh.svg")
        self.set_themed_icon(self._ui.profile_rename_button, "icons/rename.svg")
        self.set_themed_icon(self._ui.profiles_copy_name_button, "icons/copy_to_clipboard.svg")

    def _init_profile_context_menu(self):
        self.profiles_cntx_menu = QMenu()
        self._copy_profile_action = QAction("Copy profile name", self)
        self._copy_profile_action.setIcon(QIcon(get_themed_asset_icon("icons/copy_link.svg")))
        self.profiles_cntx_menu.addAction(self._copy_profile_action)
        self._copy_profile_action.triggered.connect(self.on_copy_profile_requested)

    def _load_profiles_tab(self):
        profiles_model = ProfilesModel()
        self._ui.profiles_list_view.setModel(profiles_model)
        self._ui.profiles_list_view.selectionModel().selectionChanged.connect(
            self.on_profile_selected
        )

    def on_copy_profile_requested(self):
        view_indexes = self._ui.profiles_list_view.selectedIndexes()
        if not view_indexes:
            return
        view_index = view_indexes[0]
        profile_name = view_index.data()
        QApplication.clipboard().setText(profile_name)

    def on_profile_selected(self):
        view_indexes = self._ui.profiles_list_view.selectedIndexes()
        if not view_indexes:
            return
        view_index = view_indexes[0]
        profile_name: str = view_index.data()
        self._edited_profile = profile_name
        try:
            profile_content = (self.profiles_path / profile_name).read_text()
        except Exception:
            profile_content = ""
        self._ui.profiles_text_browser.setText(profile_content)

    def on_profile_context_menu_requested(self, position):
        self.profiles_cntx_menu.exec(self._ui.profiles_list_view.mapToGlobal(position))

    def on_profile_save_file(self):
        if not self._edited_profile:
            return
        profile_name = self._edited_profile
        text = self._ui.profiles_text_browser.toPlainText()
        (self.profiles_path / profile_name).write_text(text)

    def on_profile_add(self):
        new_profile_dialog = QInputDialog(self)
        profile_name, accepted = new_profile_dialog.getText(
            self, "New profile", "Enter name:", text=""
        )
        if accepted and profile_name:
            (self.profiles_path / profile_name).touch()
            self.on_profiles_refresh()

    def on_profile_rename(self):
        view_indexes = self._ui.profiles_list_view.selectedIndexes()
        if not view_indexes:
            return
        profile_name: str = view_indexes[0].data(0)
        rename_profile_dialog = QInputDialog(self)
        new_profile_name, accepted = rename_profile_dialog.getText(
            self, "Rename profile", "Enter new name:", text=profile_name
        )
        if accepted and profile_name:
            try:
                (self.profiles_path / profile_name).rename(
                    self.profiles_path / new_profile_name
                )
            except Exception as e:
                Logger().error(f"Can't rename {profile_name}: {e}")
            self.on_profiles_refresh()

    def on_profile_remove(self):
        view_indexes = self._ui.profiles_list_view.selectedIndexes()
        if not view_indexes:
            return
        profile_name: str = view_indexes[0].data(0)
        message_box = QMessageBox(parent=self)
        message_box.setWindowTitle("Remove profile")
        message_box.setText(f"Are you sure, you want to delete the profile {profile_name}?")
        message_box.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        message_box.setIcon(QMessageBox.Icon.Question)
        reply = message_box.exec()
        if reply == QMessageBox.StandardButton.Yes:
            delete_path(self.profiles_path / profile_name)
            self.on_profiles_refresh()

    def on_profiles_refresh(self):
        profile_model: ProfilesModel = self._ui.profiles_list_view.model()  # type: ignore
        # clear selection, otherwise an old selection could remain active
        self._ui.profiles_list_view.selectionModel().clear()
        profile_model.setup_model_data()
        self._ui.profiles_list_view.repaint()

    # Remote

    def _load_remotes_tab(self):
        self._remotes_controller.update()

    def _init_remotes_tab(self):
        self._ui.remote_refresh_button.clicked.connect(self._remotes_controller.update)
        self.set_themed_icon(self._ui.remote_refresh_button, "icons/refresh.svg")
        self._ui.remote_login_button.clicked.connect(self.on_remotes_login)
        self.set_themed_icon(self._ui.remote_login_button, "icons/login.svg")
        self._ui.remote_toggle_disabled_button.clicked.connect(
            self._remotes_controller.remote_disable
        )
        self.set_themed_icon(self._ui.remote_toggle_disabled_button, "icons/hide.svg")
        self._ui.remote_add_button.clicked.connect(self.on_remote_add)
        self.set_themed_icon(self._ui.remote_add_button, "icons/plus_rounded.svg")
        self._ui.remote_remove_button.clicked.connect(self.on_remote_remove)
        self.set_themed_icon(self._ui.remote_remove_button, "icons/delete.svg")

        self._ui.remote_move_up_button.clicked.connect(self._remotes_controller.move_up)
        self.set_themed_icon(self._ui.remote_move_up_button, "icons/arrow_up.svg")
        self._ui.remote_move_down_button.clicked.connect(self._remotes_controller.move_down)
        self.set_themed_icon(self._ui.remote_move_down_button, "icons/arrow_down.svg")

        self._ui.remote_move_top_button.clicked.connect(self._remotes_controller.move_to_top)
        self.set_themed_icon(self._ui.remote_move_top_button, "icons/expand_less.svg")
        self._ui.remote_move_bottom_button.clicked.connect(
            self._remotes_controller.move_to_bottom
        )
        self.set_themed_icon(self._ui.remote_move_bottom_button, "icons/expand.svg")

        self._ui.remotes_edit_button.clicked.connect(self.on_remote_edit)
        self.set_themed_icon(self._ui.remotes_edit_button, "icons/edit.svg")
        self._ui.remotes_tree_view.doubleClicked.connect(self.on_remote_edit)

        self._ui.remotes_tree_view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._ui.remotes_tree_view.customContextMenuRequested.connect(
            self.on_remote_context_menu_requested
        )
        self._init_remote_context_menu()

    def on_remote_context_menu_requested(self, position):
        self._remotes_cntx_menu.exec(self._ui.remotes_tree_view.mapToGlobal(position))

    def _init_remote_context_menu(self):
        self._remotes_cntx_menu = QMenu()
        self._copy_remote_action = QAction("Copy remote name", self)
        self._copy_remote_action.setIcon(QIcon(get_themed_asset_icon("icons/copy_link.svg")))
        self._remotes_cntx_menu.addAction(self._copy_remote_action)
        self._copy_remote_action.triggered.connect(self._remotes_controller.copy_remote_name)

        self._edit_remote_action = QAction("Edit remote", self)
        self._edit_remote_action.setIcon(QIcon(get_themed_asset_icon("icons/edit.svg")))
        self._remotes_cntx_menu.addAction(self._edit_remote_action)
        self._edit_remote_action.triggered.connect(self.on_remote_edit)

        self._add_remote_action = QAction("Add new remote", self)
        self._add_remote_action.setIcon(QIcon(get_themed_asset_icon("icons/plus_rounded.svg")))
        self._remotes_cntx_menu.addAction(self._add_remote_action)
        self._add_remote_action.triggered.connect(self.on_remote_add)

        self._remove_remote_action = QAction("Remove remote", self)
        self._remove_remote_action.setIcon(QIcon(get_themed_asset_icon("icons/delete.svg")))
        self._remotes_cntx_menu.addAction(self._remove_remote_action)
        self._remove_remote_action.triggered.connect(self.on_remote_remove)

        self._disable_profile_action = QAction("Disable/Enable remote", self)
        self._disable_profile_action.setIcon(QIcon(get_themed_asset_icon("icons/hide.svg")))
        self._remotes_cntx_menu.addAction(self._disable_profile_action)
        self._disable_profile_action.triggered.connect(self._remotes_controller.remote_disable)

        self._login_remotes_action = QAction("(Multi)Login to remote", self)
        self._login_remotes_action.setIcon(QIcon(get_themed_asset_icon("icons/login.svg")))
        self._remotes_cntx_menu.addAction(self._login_remotes_action)
        self._login_remotes_action.triggered.connect(self.on_remotes_login)

    def on_remote_edit(self, model_index):
        remote_item = self._remotes_controller.get_selected_remote()
        if not remote_item:
            return
        self.remote_edit_dialog = RemoteEditDialog(remote_item, self._remotes_controller, self)
        self.remote_edit_dialog.exec()

    def on_remotes_login(self):
        remote_item = self._remotes_controller.get_selected_remote()
        if not remote_item:
            return
        remotes = app.conan_api.get_remotes_from_same_server(remote_item)
        if not remotes:
            return
        self.remote_login_dialog = RemoteLoginDialog(remotes, self._remotes_controller, self)
        self.remote_login_dialog.exec()

    def on_remote_add(self, model_index):
        self.remote_edit_dialog = RemoteEditDialog(None, self._remotes_controller, self)
        self.remote_edit_dialog.exec()

    def on_remote_remove(self, model_index):
        self._remotes_controller.on_remove()

    # Editables tab

    def _load_editables_tab(self):
        self._ui.editables_add_button.clicked.connect(self.on_editable_add)
        self._ui.editables_remove_button.clicked.connect(self.on_editable_remove)
        self._ui.editables_refresh_button.clicked.connect(self._editable_controller.update)
        self._ui.editables_edit_button.clicked.connect(self.on_editable_edit)
        self._ui.editables_ref_view.doubleClicked.connect(self.on_editable_edit)

        self.set_themed_icon(self._ui.editables_add_button, "icons/plus_rounded.svg")
        self.set_themed_icon(self._ui.editables_remove_button, "icons/delete.svg")
        self.set_themed_icon(self._ui.editables_refresh_button, "icons/refresh.svg")
        self.set_themed_icon(self._ui.editables_edit_button, "icons/edit.svg")

    def on_editable_add(self, model_index):
        self.remote_edit_dialog = EditableEditDialog(None, self._editable_controller, self)
        self.remote_edit_dialog.exec()

    def on_editable_edit(self, model_index):
        editable = self._editable_controller.get_selected_editable()
        if not editable:
            return
        self.remote_edit_dialog = EditableEditDialog(editable, self._editable_controller, self)
        self.remote_edit_dialog.exec()

    def on_editable_remove(self, model_index):
        editable_item = self._editable_controller.get_selected_editable()
        if not editable_item:
            return
        message_box = QMessageBox(parent=self)
        message_box.setWindowTitle("Remove editable")
        message_box.setText(
            "Are you sure, you want to delete the editable" + f"{editable_item.conan_ref}?"
        )
        standard_button = QMessageBox.StandardButton
        message_box.setStandardButtons(standard_button.Yes | standard_button.No)
        message_box.setIcon(QMessageBox.Icon.Question)
        reply = message_box.exec()
        if reply == standard_button.Yes:
            self._editable_controller.remove(editable_item)

    # Conan Config

    def _load_config_file_tab(self):
        self.set_themed_icon(self._ui.config_file_save_button, "icons/save.svg")
        try:
            self._ui.config_file_text_browser.setText(
                app.conan_api.get_config_file_path().read_text()
            )
            self._ui.config_file_save_button.clicked.connect(self.on_save_config_file)
        except Exception:
            Logger().error("Cannot read Conan config file!")

    def on_save_config_file(self):
        app.conan_api.get_config_file_path().write_text(
            self._ui.config_file_text_browser.toPlainText()
        )
        # restart conan api to apply changes internally
        app.conan_api.init_api()
        Logger().info("Applying Changes to Conan...")
        # re-init info tab to show the changes
        self._load_info_tab()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'plugins.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QHBoxLayout, QHeaderView,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QTreeView, QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(489, 333)
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 1, 1, 1)
        self.buttons_frame = QFrame(Form)
        self.buttons_frame.setObjectName(u"buttons_frame")
        self.buttons_frame.setFrameShape(QFrame.StyledPanel)
        self.buttons_frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.buttons_frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, -1, 0, -1)
        self.add_plugin_button = QPushButton(self.buttons_frame)
        self.add_plugin_button.setObjectName(u"add_plugin_button")
        icon = QIcon()
        icon.addFile(u"../../assets/icons/plus_rounded.png", QSize(), QIcon.Mode.Normal, QIcon.State.On)
        self.add_plugin_button.setIcon(icon)

        self.horizontalLayout.addWidget(self.add_plugin_button)

        self.remove_plugin_button = QPushButton(self.buttons_frame)
        self.remove_plugin_button.setObjectName(u"remove_plugin_button")
        icon1 = QIcon()
        icon1.addFile(u"../../assets/icons/minus_rounded.png", QSize(), QIcon.Mode.Normal, QIcon.State.On)
        self.remove_plugin_button.setIcon(icon1)

        self.horizontalLayout.addWidget(self.remove_plugin_button)

        self.reload_plugin_button = QPushButton(self.buttons_frame)
        self.reload_plugin_button.setObjectName(u"reload_plugin_button")

        self.horizontalLayout.addWidget(self.reload_plugin_button)

        self.buttons_spacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.buttons_spacer)


        self.verticalLayout.addWidget(self.buttons_frame)

        self.path_label = QLabel(Form)
        self.path_label.setObjectName(u"path_label")

        self.verticalLayout.addWidget(self.path_label)

        self.plugins_tree_view = QTreeView(Form)
        self.plugins_tree_view.setObjectName(u"plugins_tree_view")
        self.plugins_tree_view.setFrameShape(QFrame.NoFrame)

        self.verticalLayout.addWidget(self.plugins_tree_view)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.add_plugin_button.setText(QCoreApplication.translate("Form", u"Add new Plugin", None))
        self.remove_plugin_button.setText(QCoreApplication.translate("Form", u"Remove Plugin", None))
        self.reload_plugin_button.setText(QCoreApplication.translate("Form", u"Reload Plugin", None))
        self.path_label.setText("")
    # retranslateUi




from threading import Thread
from typing import Callable, List

from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QCompleter, QLineEdit, QListView
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanPkgRef, ConanRef
from conan_explorer.ui.common.theming import get_gui_dark_mode


class ConanRefLineEdit(QLineEdit):
    """Adds completions for Conan references and a validator."""

    completion_finished = Signal()
    MINIMUM_CHARS_FOR_QUERY = 6
    INVALID_COLOR = "LightCoral"
    VALID_COLOR_LIGHT = "#37efba"  # light green
    VALID_COLOR_DARK = "#007b50"  # dark green

    def __init__(self, parent, validator_enabled=True):
        super().__init__(parent)
        self.validator_enabled = validator_enabled
        self.is_valid = False
        completer = QCompleter([], self)
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        completer.setMaxVisibleItems(10)
        completer.setModelSorting(QCompleter.ModelSorting.CaseInsensitivelySortedModel)
        completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)

        self._first_show = True  # don't call completer on first show
        self._completion_thread = None
        self._loading_cbk = None
        completer_popup = QListView(self)
        completer_popup.setLayoutMode(QListView.LayoutMode.Batched)
        completer_popup.setUniformItemSizes(True)
        completer.setPopup(completer_popup)
        self.setCompleter(completer)

        self.completion_finished.connect(self.completer().complete)
        self.textChanged.connect(self.on_text_changed)

    @override
    def setEnabled(self, enabled: bool):
        # apply grey color to text, when disabled manually
        if enabled:
            self.setStyleSheet("")  # remove grey color
        else:
            self.setStyleSheet("color: grey;")
        super().setEnabled(enabled)

    @override
    def showEvent(self, event):
        self.load_completion_refs()
        if self._first_show:
            self.completer().popup().hide()
            self._first_show = True
        super().showEvent(event)

    def load_completion_refs(self):
        combined_refs = set()
        combined_refs.update(app.conan_api.info_cache.get_all_local_refs())
        combined_refs.update(app.conan_api.info_cache.get_all_remote_refs())
        self.completer().model().setStringList(sorted(combined_refs, reverse=True))  # type: ignore

    def cleanup(self):
        if self._completion_thread:
            self._completion_thread.join(1)

    def set_loading_callback(self, loading_cbk: Callable):
        self._loading_cbk = loading_cbk

    def on_text_changed(self, text: str):
        self.validate(text)
        self.search_query(text)

    def validate(self, conan_ref: str):
        """Validate ConanRef or PackageReference. Empty text is invalid."""
        if not self.validator_enabled:
            return
        if not conan_ref:
            self.is_valid = False
        else:
            try:
                if ":" in conan_ref:
                    ConanPkgRef.loads(conan_ref)
                else:
                    ConanRef.loads(conan_ref)
                self.is_valid = True
            except Exception:
                self.is_valid = False

        # color background
        if self.is_valid:
            if get_gui_dark_mode():
                self.setStyleSheet(f"background: {self.VALID_COLOR_DARK};")
            else:
                self.setStyleSheet(f"background: {self.VALID_COLOR_LIGHT};")
        else:  # if it does error it's invalid format, thus red
            self.setStyleSheet(f"background: {self.INVALID_COLOR};")

    def search_query(self, conan_ref: str):
        if len(conan_ref) < self.MINIMUM_CHARS_FOR_QUERY:  # skip searching for such broad terms
            return
        # start a query for all similar packages with conan search by starting a new thread
        if (
            self._completion_thread and self._completion_thread.is_alive()
        ):  # one query at a time
            return
        self._completion_thread = Thread(
            target=self.load_completion,
            args=[
                conan_ref,
            ],
        )
        self._completion_thread.start()
        if self._loading_cbk:
            self._loading_cbk()

    def load_completion(self, text: str):
        if (
            any(
                [
                    entry.startswith(text)
                    for entry in app.conan_api.info_cache.get_all_remote_refs()
                ]
            )
            or self.is_valid
        ):
            return
        recipes = app.conan_api.search_recipes_in_remotes(f"{text}*")  # can take very long time
        if app.conan_api:  # program can shut down and conan_api destroyed
            try:
                app.conan_api.info_cache.update_remote_package_list(recipes)  # add to cache
                remote_refs = app.conan_api.info_cache.get_all_remote_refs()
                # add two list together -> filter is applied later
                current_completions: List[str] = self.completer().model().stringList()  # type: ignore
                new_completions = set(remote_refs + current_completions)
                if len(new_completions) > len(current_completions):
                    self.completer().model().setStringList(sorted(new_completions))  # type: ignore
                    self.completion_finished.emit()

            except Exception as e:
                Logger().error(f"Failed load completion: {str(e)}")



from typing import Callable, Optional

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QKeySequence, QShortcut
from PySide6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from conan_explorer.app.logger import Logger
from conan_explorer.ui.common import ThemedWidget
from conan_explorer.ui.widgets.toggle import AnimatedToggle

from .fluent_window import RIGHT_MENU_MAX_WIDTH, gen_obj_name


class SideSubMenu(ThemedWidget):
    def __init__(
        self, parent_stacked_widget: QStackedWidget, title: str = "", is_top_level=False
    ):
        super().__init__(parent_stacked_widget)
        from .side_menu_ui import Ui_SideMenu  # need to resolve circular import

        self.ui = Ui_SideMenu()
        self.ui.setupUi(self)

        self.parent_stacked_widget = parent_stacked_widget
        self.parent_stacked_widget.addWidget(self)
        self.title = title
        self.is_top_level = is_top_level
        self.set_title(title)
        self._content_layout = self.ui.content_frame_layout
        self.set_themed_icon(self.ui.side_menu_title_button, "icons/back.svg")

        if is_top_level:
            self.ui.side_menu_title_button.hide()  # off per default

    def reset_widgets(self):
        while self._content_layout.count() > 1:
            widget = self._content_layout.takeAt(0)
            if widget in [None, self.ui.side_menu_spacer]:
                continue
            self._content_layout.removeItem(widget)

    def set_title(self, title: str):
        self.ui.side_menu_title_label.setText(title)

    def enable_collapsible(self) -> bool:
        """
        Enable this side menu being collapsed. The side_menu_title_button will be used, so this
        must be a top level menu, otherwise the back button could not be operated anymore.
        """
        if not self.is_top_level:
            return False
        self.set_themed_icon(self.ui.side_menu_title_button, "icons/expand.svg")
        self.ui.side_menu_title_button.clicked.connect(
            self.on_expand_minimize
        )  # off per default
        return True

    def on_expand_minimize(self):
        """The title button can be used to minimize a submenu"""
        if self.ui.side_menu_content_frame.height() > 0:
            self.ui.side_menu_content_frame.setMaximumHeight(0)
            self.set_themed_icon(self.ui.side_menu_title_button, "icons/forward.svg")
        else:
            self.set_themed_icon(self.ui.side_menu_title_button, "icons/expand.svg")
            self.ui.side_menu_content_frame.setMaximumHeight(4096)

    def get_menu_entry_by_name(self, name: str) -> Optional[QWidget]:
        return self.findChild(QWidget, gen_obj_name(name))  # type:ignore

    def add_custom_menu_entry(self, widget: QWidget, name: Optional[str] = None):
        """Very basic custom entry, no extra functions"""
        if name:
            widget.setObjectName(gen_obj_name(name))
        self._content_layout.insertWidget(self._content_layout.count() - 1, widget)

    def add_menu_line(self):
        line = QFrame(self)
        line.setLineWidth(2)
        line.setMidLineWidth(0)
        line.setFrameShape(QFrame.Shape.HLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)
        self.add_custom_menu_entry(line, "line")

    def add_named_custom_entry(
        self, name: str, widget: QWidget, asset_icon: str = "", force_v_layout=False
    ):
        "Creates a Frame with a text label and a custom widget under it and adds it to the menu"
        label = QLabel(text=name, parent=self)
        label.adjustSize()  # adjust layout according to size and throw a warning, if too big?
        label.setObjectName(gen_obj_name(name) + "_label")
        label.setMaximumHeight(30)
        size_policy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        label.setSizePolicy(size_policy)
        icon = None
        if asset_icon:
            icon = QLabel(parent=self)
            icon.setObjectName(gen_obj_name(name) + "_icon")
            self.set_themed_icon(icon, asset_icon)
        widget.adjustSize()
        widget.setMinimumHeight(50)
        widget.setMaximumHeight(100)
        widget.setObjectName(gen_obj_name(name) + "_widget")

        frame = QFrame(self)
        if force_v_layout or label.width() > (
            RIGHT_MENU_MAX_WIDTH - widget.width() - 30
        ):  # aggressive 30 px padding
            layout = QVBoxLayout(frame)
            frame.setLayout(layout)
            if icon is not None:  # in vmode the icon still needs to be placed in the
                # same row, so we need an extra h-layout
                horizontal_layout = QHBoxLayout()
                horizontal_layout.addWidget(icon)
                horizontal_layout.addWidget(label)
                horizontal_layout.addSpacerItem(
                    QSpacerItem(
                        40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
                    )
                )
                layout.addLayout(horizontal_layout)
                layout.setStretch(1, 1)
            else:
                layout.addWidget(label)
        else:
            layout = QHBoxLayout(frame)
            frame.setLayout(layout)
            if icon is not None:
                layout.addWidget(icon)
            layout.addWidget(label)
            layout.setStretch(1, 1)

        if label.width() > RIGHT_MENU_MAX_WIDTH:
            Logger().debug("%s right side menu exceeds max width!", name)
        frame.layout().setContentsMargins(5, 0, 5, 0)
        frame.layout().setSpacing(4)

        frame.layout().addWidget(widget)
        self.add_custom_menu_entry(frame, name)

    def add_toggle_menu_entry(
        self, name: str, target: Callable, initial_state: bool, asset_icon: str = ""
    ):
        toggle = AnimatedToggle(self)
        toggle.setChecked(initial_state)
        toggle.stateChanged.connect(target)
        self.add_named_custom_entry(name, toggle, asset_icon, force_v_layout=True)
        return toggle

    def add_sub_menu(self, sub_menu: "SideSubMenu", asset_icon: str = ""):
        button = self.add_button_menu_entry(
            sub_menu.title, sub_menu.ui.side_menu_title_button.show, asset_icon
        )
        button.clicked.connect(lambda: self.parent_stacked_widget.setCurrentWidget(sub_menu))
        sub_menu.ui.side_menu_title_button.clicked.connect(
            lambda: self.parent_stacked_widget.setCurrentWidget(self)
        )
        sub_menu.setObjectName(gen_obj_name(sub_menu.title) + "_widget")
        return button

    def add_button_menu_entry(
        self,
        name: str,
        target: Callable,
        asset_icon: str = "",
        shortcut: Optional[QKeySequence] = None,
        shortcut_parent: Optional[QWidget] = None,
    ):
        """Adds a button with an icon and links with a callable.
        Optionally can have a key shortcut."""
        button = QPushButton(self)
        button.setMinimumSize(QSize(64, 50))
        button.setMaximumHeight(50)
        if asset_icon:
            self.set_themed_icon(button, asset_icon)
        button.setText(name)
        button.setStyleSheet("text-align:left")
        # insert before spacer
        self.add_custom_menu_entry(button, name)

        button.clicked.connect(target)

        if not shortcut:
            return button
        assert shortcut_parent, "Add shortcut_parent if shortcut is True!"

        # use global shortcut instead of button.setShortcut -> Works from anywhere
        shortcut_obj = QShortcut(shortcut, shortcut_parent)
        shortcut_obj.activated.connect(target)
        button.setText(f"{button.text()} ({shortcut.toString()})")
        shortcut_obj.setContext(Qt.ShortcutContext.ApplicationShortcut)

        return button
